"""
Utilities
"""
import csv
import datetime
from itertools import chain
import json
import logging
import os
import signal
from typing import Callable
from typing import Iterator
from typing import List
from typing import Sequence
from typing import Set
from typing import TypeVar
import warnings

import networkx as nx
import yaml


try:
    _ = signal.SIGALRM
    _HAS_SIGALRM = True
except AttributeError:
    _HAS_SIGALRM = False
_Template = TypeVar("_Template")

ENCODING = "UTF-8"


def get_logging_level() -> int:
    """
    Detect current logging level
    """
    return logging.getLogger().getEffectiveLevel()


def silence_third_party():
    """
    Silence logging of third party packages
    """
    warnings.filterwarnings("ignore", category=UserWarning)
    logging.getLogger("pytorch_lightning").setLevel(logging.ERROR)


class _LoggingWrapper:
    def __init__(self, logging_level: int, fun: Callable[..., _Template]):
        self._logging_level = logging_level
        self._fun = fun

    def __call__(self, *args, **kwargs):
        logging.basicConfig(level=self._logging_level)
        if self._logging_level > logging.DEBUG:
            silence_third_party()
        return self._fun(*args, **kwargs)


def require_logging(fun: Callable[..., _Template]) -> Callable[..., _Template]:
    """
    Set logging level in a new process
    """
    logging_level = get_logging_level()
    return _LoggingWrapper(logging_level=logging_level, fun=fun)


class Timer:
    """
    Record duration as a context manager
    """

    def __init__(self, name: str):
        self._name = name
        class_name = f"{self.__class__.__module__}.{self.__class__.__name__}"
        self._logger = logging.getLogger(class_name)

        self._start: datetime.datetime = None

    @property
    def start(self) -> datetime.datetime:
        """
        Start time
        """
        return self._start

    @property
    def duration(self) -> datetime.timedelta:
        """
        Elapsed time since start
        """
        return datetime.datetime.now() - self._start

    def __enter__(self):
        self._start = datetime.datetime.now()
        self._logger.info("%s starts at %s", self._name, self._start)
        return self

    def __exit__(self, *_):
        self._logger.info(
            "Duration of %s: %.6f", self._name, self.duration.total_seconds()
        )


class Timeout:
    """
    Control execution duration as a context manager

    Stolen from the following answer, which is under the "CC BY-SA 4.0" license
    https://stackoverflow.com/a/39773925

    Exemples:
    >>> import time
    >>> with Timeout(seconds=3):  # doctest: +SKIP
    ...     time.sleep(4)
    Traceback (most recent call last):
        ...
    TimeoutError
    """

    def __init__(self, seconds: int = 0):
        if not _HAS_SIGALRM:
            warnings.warn(
                "signal.SIGALRM is not supported and "
                f"{Timeout.__module__}.{Timeout.__name__} cannot work"
            )
        self._seconds = int(seconds)

    def _handle_timeout(self, *_):
        # pylint: disable=no-self-use
        raise TimeoutError

    def __enter__(self):
        if _HAS_SIGALRM:
            signal.signal(signal.SIGALRM, self._handle_timeout)
            signal.alarm(self._seconds)

    def __exit__(self, *_):
        if _HAS_SIGALRM:
            signal.alarm(0)


def dump_csv(filename: str, data: Sequence[Sequence], headers: Sequence[str] = None):
    """
    Dump data into a csv file
    """
    with open(filename, "w", encoding=ENCODING, newline="") as obj:
        writer = csv.writer(obj)
        if headers is not None:
            writer.writerow(headers)
        writer.writerows(data)


def load_csv(filename: str):
    """
    Load data from a csv file
    """
    with open(filename, encoding=ENCODING) as obj:
        reader = csv.reader(obj)
        for row in reader:
            yield row


def dump_json(filename: str, data):
    """
    Dump data into a json file
    """
    with open(filename, "w", encoding=ENCODING) as obj:
        json.dump(data, obj, ensure_ascii=False, indent=2)


def load_json(filename: str):
    """
    Load data from a json file
    """
    with open(filename, encoding=ENCODING) as obj:
        return json.load(obj)


class YamlComposeLoader(yaml.SafeLoader):
    # pylint: disable=too-many-ancestors
    """
    Equip yaml.SafeLoader with the "!include" tag
    """

    @classmethod
    def load(cls, filename: str) -> dict:
        """
        Load from file
        """
        with open(filename, encoding=ENCODING) as obj:
            return yaml.load(obj, cls)

    @staticmethod
    def include(loader: yaml.Loader, node: yaml.Node):
        """
        Handle the "!include" tag
        """
        filename = os.path.join(
            os.path.dirname(loader.name), loader.construct_scalar(node)
        )
        return YamlComposeLoader.load(filename)


YamlComposeLoader.add_constructor("!include", YamlComposeLoader.include)


def topological_sort(
    nodes: Set[_Template],
    predecessors: Callable[[_Template], Iterator[_Template]],
    successors: Callable[[_Template], Iterator[_Template]],
) -> List[Set[_Template]]:
    """
    Sort nodes with predecessors first
    """
    graph = {node: set(successors(node)) for node in nodes}
    components = list(nx.strongly_connected_components(nx.DiGraph(graph)))
    node2component = {
        node: index for index, component in enumerate(components) for node in component
    }
    super_graph = {
        index: {node2component[child] for node in component for child in graph[node]}
        - {index}
        for index, component in enumerate(components)
    }
    return [
        set(chain(*[components[index] for index in layer]))
        for layer in nx.topological_generations(nx.DiGraph(super_graph))
    ]



"""
Define the interface for algorithms to access data
"""
from abc import ABC
from datetime import timedelta
from typing import Dict
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd

from .graph import Node


class DataLoader(ABC):
    """
    The abstract interface to access data
    """

    @property
    def entities(self) -> Sequence[str]:
        """
        Fetch names of available entities
        """
        raise NotImplementedError

    @property
    def metrics(self) -> Dict[str, Sequence[str]]:
        """
        Fetch available metrics as a mapping from entity names to metric names
        """
        raise NotImplementedError

    @property
    def nodes(self) -> Sequence[Node]:
        """
        Pairs of entity and metric
        """
        return [
            Node(entity=entity, metric=metric)
            for entity, metrics in self.metrics.items()
            for metric in metrics
        ]

    def load(
        self,
        entity: str,
        metric: str,
        start: float,
        end: float,
        interval: timedelta,
        **kwargs,
    ) -> Union[None, Sequence[float]]:
        # pylint: disable=too-many-arguments
        """
        Load the time series for the given metric of the given entity

        start: expected start time of the time series,
            which is a unix timestamp in seconds
        end: expected end time of the time series,
            which is a unix timestamp in seconds
        interval: interval between two data points
        """
        raise NotImplementedError

    @staticmethod
    def preprocess(
        time_series: Sequence[Tuple[float, float]],
        start: float,
        end: float,
        interval: timedelta,
        **kwargs,
    ) -> Sequence[float]:
        """
        Truncate the time series and fill missing data points

        This method is edited from k-Shape, which is released under the MIT license.
        See https://github.com/sieve-microservices/kshape/blob/master/README.md

        time_series: 2-tuples (timestamp, value),
            where timestamp is a unix timestamp in seconds
        start: expected start time of the time series,
            which is a unix timestamp in seconds
        end: expected end time of the time series,
            which is a unix timestamp in seconds
        interval: interval between two data points
        """
        if not time_series:
            return None
        data: np.ndarray = np.array(time_series)
        # 1. Truncate the time series and make sure that [start, end] is the boundry
        data = data[(data[:, 0] >= start) & (data[:, 0] <= end), :]
        if len(data) == 0:
            return None
        data = np.vstack(
            [
                data,
                np.array([(start, np.nan), (end, np.nan)]),
            ]
        )

        # 2. Fill missing data points with fixed frequency
        data_frame = pd.DataFrame(data)
        data_frame[0] = pd.to_datetime(
            data_frame[0], unit=kwargs.get("unit", "s"), utc=True
        )
        data_frame: pd.DataFrame = (
            data_frame.set_index(0).resample(interval, origin="start").mean()
        )
        data_frame.interpolate(method="time", limit_direction="both", inplace=True)
        data_frame.fillna(method="bfill", inplace=True)

        # 3. Return values only
        data_frame.sort_index(inplace=True)
        return tuple(data_frame[1])


class MemoryDataLoader(DataLoader):
    """
    Implement DataLoader with data in memory
    """

    def __init__(self, data: Dict[str, Dict[str, Sequence[Tuple[float, float]]]]):
        """
        data: data[entity][metric] is the 2-tuples (timestamp, value)
            for the given metric of the given entity
        """
        self._data = data

    @property
    def entities(self) -> Sequence[str]:
        return tuple(self._data.keys())

    @property
    def metrics(self) -> Dict[str, Sequence[str]]:
        return {entity: tuple(metrics.keys()) for entity, metrics in self._data.items()}

    def load(
        self,
        entity: str,
        metric: str,
        start: float,
        end: float,
        interval: timedelta,
        **kwargs,
    ) -> Union[None, Sequence[float]]:
        # pylint: disable=too-many-arguments
        if entity not in self._data or metric not in self._data[entity]:
            return None
        return self.preprocess(
            time_series=self._data[entity][metric],
            start=start,
            end=end,
            interval=interval,
            **kwargs,
        )



"""
Robustness evaluation
"""
from enum import Enum
import logging
import os
import pickle
from typing import Dict
from typing import List

from . import _SLI
from . import SimCase
from . import SimDataset
from ..comparison import IDEAL_METHOD
from ..comparison import report_evaluation
from ...alg.common import Evaluation
from ...model.case import Case
from ...model.graph import Node
from ...utils import dump_csv
from ...utils import load_json


class Intensity(Enum):
    """
    Strong dependency intensity leads to unobvious root cause,
    dealing with which shows strong robustness.
    """

    WEAK = "weak"
    STRONG = "strong"
    MIXED = "mixed"

    @staticmethod
    def classify(data_dir: str) -> "Intensity":
        """
        Classify a case
        """
        detail_filename = os.path.join(data_dir, SimCase.DETAIL_FILENAME)
        with open(detail_filename, "rb") as obj:
            data = pickle.load(obj)
        ratio = abs(data["stds"] * data["weight"][_SLI, :] / data["stds"][_SLI])
        root_cause = data["fault"] != 0
        if not any((ratio <= 1) & root_cause):
            return Intensity.STRONG
        if any((ratio > 1) & root_cause):
            return Intensity.MIXED
        return Intensity.WEAK


def _evaluate_graph(
    cases: List[Case],
    reports: Dict[Intensity, Dict[str, Evaluation]],
    cache_dir: str,
    data_dir: str,
):
    """
    Evaluate for a single graph
    """
    answer_groups: Dict[str, List[Case]] = {intensity: [] for intensity in Intensity}
    index_groups: Dict[Intensity, List[int]] = {
        intensity: [] for intensity in Intensity
    }
    for i, case in enumerate(cases):
        intensity = Intensity.classify(
            data_dir=os.path.join(data_dir, SimDataset.CASES_FOLDER, str(i))
        )
        answer_groups[intensity].append(case.answer)
        index_groups[intensity].append(i)

    for filename in os.listdir(cache_dir):
        if not filename.endswith(".json"):
            continue
        method = filename[:-5]
        ranks = [
            [Node(entity=node["entity"], metric=node["metric"]) for node in rank]
            for rank in load_json(os.path.join(cache_dir, filename))
        ]
        for intensity, answers in answer_groups.items():
            if method not in reports[intensity]:
                reports[intensity][method] = Evaluation()
            for i, answer in zip(index_groups[intensity], answers):
                reports[intensity][method](ranks=ranks[i], answers=answer)

    for intensity, answers in answer_groups.items():
        if IDEAL_METHOD not in reports[intensity]:
            reports[intensity][IDEAL_METHOD] = Evaluation()
        for answer in answers:
            reports[intensity][IDEAL_METHOD](ranks=list(answer), answers=answer)
    return reports


def evaluate(num_graph: int, cache_dir: str, data_dir: str, report_dir: str):
    """
    Classify cases based on dependency intensities and evaluate separately
    """
    os.makedirs(report_dir, exist_ok=True)
    reports: Dict[Intensity, Dict[str, Evaluation]] = {
        intensity: {} for intensity in Intensity
    }

    logger = logging.getLogger(evaluate.__module__)
    for i in range(num_graph):
        cases_dir = os.path.join(data_dir, str(i))
        logger.info("Loading from %s", cases_dir)
        cases = SimDataset.load(folder=cases_dir).cases
        reports = _evaluate_graph(
            cases=cases,
            reports=reports,
            cache_dir=os.path.join(cache_dir, str(i)),
            data_dir=cases_dir,
        )

    for intensity, sub_reports in reports.items():
        dump_csv(
            filename=os.path.join(report_dir, f"{intensity.value}.csv"),
            data=[
                report_evaluation(method, report)
                for method, report in sub_reports.items()
            ],
            headers=["method", "AC@1", "AC@3", "AC@5", "Avg@5"],
        )
    dump_csv(
        filename=os.path.join(report_dir, "count.csv"),
        data=[
            [intensity.value, list(sub_reports.values())[0].num]
            for intensity, sub_reports in reports.items()
        ],
        headers=["intensity", "count"],
    )



"""
Anomaly Detection in Streams with Extreme Value Theory

See SPOT in KDD'17
https://doi.org/10.1145/3097983.3098144
"""
from typing import Dict
from typing import List
from typing import Sequence

from ads_evt import biSPOT
import numpy as np

from .base import Score
from .common import DecomposableScorer
from ..model.case import CaseData
from ..model.graph import Graph
from ..model.graph import Node


def spot(train_y: np.ndarray, test_y: np.ndarray, proba: float = 1e-4) -> np.ndarray:
    """
    Estimate to what extend each value in test_y violates
    the normal distribution defined by train_y
    """
    model = biSPOT(q=proba)
    model.fit(init_data=train_y, data=test_y)
    model.initialize()
    results = model.run(with_alarm=False)
    scores: List[float] = []
    for index, (upper, lower) in enumerate(
        zip(results["upper_thresholds"], results["lower_thresholds"])
    ):
        width: float = upper - lower
        if width <= 0:
            width = 1
        if test_y[index] > upper:
            scores.append((test_y[index] - upper) / width)
        elif test_y[index] < lower:
            scores.append((lower - test_y[index]) / width)
        else:
            scores.append(0)

    return np.array(scores)


class SPOTScorer(DecomposableScorer):
    """
    Score nodes by SPOT
    """

    def __init__(self, proba: float = 1e-4, **kwargs):
        super().__init__(**kwargs)
        self._porba = proba

    def score_node(
        self,
        graph: Graph,
        series: Dict[Node, Sequence[float]],
        node: Node,
        data: CaseData,
    ) -> Score:
        series_y = np.array(series[node])
        train_y: np.ndarray = series_y[: data.train_window]
        test_y: np.ndarray = series_y[-data.test_window :]
        scores = spot(train_y, test_y, proba=self._porba)
        spot_score = self._aggregator(abs(scores))
        if spot_score == 0:
            return None
        score = Score(spot_score)
        score["spot-score"] = spot_score
        return score



"""
Ranking based on the broken edges in the invariant network model
"""
from typing import Dict

import numpy as np

from .crd import CRD
from .enmf import ENMF
from .enmf import ENMFSoft
from ..base import Score
from ..base import Scorer
from ...model.case import CaseData
from ...model.graph import Graph
from ...model.graph import Node


class ENMFScorer(Scorer):
    """
    Score based on the ENMF model
    """

    def __init__(
        self,
        model_params: dict = None,
        use_softmax: bool = False,
        discrete: bool = True,
        **kwargs
    ):
        super().__init__(**kwargs)
        if model_params is None:
            model_params = {}
        if use_softmax:
            self._model = ENMFSoft(**model_params)
        else:
            self._model = ENMF(**model_params)
        self._discrete = discrete

    def score(
        self,
        graph: Graph,
        data: CaseData,
        current: float,
        scores: Dict[Node, Score] = None,
    ) -> Dict[Node, Score]:
        series = data.load_data(graph, current)
        nodes = list(series.keys())
        time_series = np.array([series[node] for node in nodes]).T
        self._model.train(time_series[: data.train_window, :])
        broken_scores = self._model.rank(
            data=time_series[-data.test_window - self._model.start :, :],
            discrete=self._discrete,
        )
        results = {
            node: Score(score=broken_score)
            for broken_score, node in zip(broken_scores, nodes)
        }
        if scores is None:
            return results
        return {node: score.update(results[node]) for node, score in scores.items()}


class CRDScorer(ENMFScorer):
    """
    Score based on the CRD model
    """

    def __init__(self, model_params: dict = None, discrete: bool = True, **kwargs):
        super().__init__(discrete=discrete, **kwargs)
        if model_params is None:
            model_params = {}
        model_params = dict(cuda=self._cuda, seed=self._seed, **model_params)
        self._model = CRD(**model_params)



"""
Correlation
"""
from typing import Dict
from typing import Sequence
from typing import Set

import pandas as pd
import pingouin as pg
from scipy.stats import pearsonr

from .base import Score
from .common import DecomposableScorer
from ..model.case import CaseData
from ..model.graph import Graph
from ..model.graph import Node


def partial_correlation(
    node: Node,
    cause: Node,
    graph: Graph,
    series: Dict[Node, Sequence[float]],
    corr_type: str = "pearson",
) -> float:
    """
    Partial correlation coefficient

    corr_type: For the "method" parameter of pingouin.partial_corr
    """
    if node == cause:
        return 1

    confounders = graph.parents(node) | graph.parents(cause)
    confounders -= {node, cause}
    data_frame = pd.DataFrame(series)

    nodes: Set[Node] = set(data_frame.index)
    confounders = {
        confounder
        for confounder in confounders
        if confounder in nodes and len(data_frame[confounder].unique()) > 1
    }

    return pg.partial_corr(
        data=data_frame, x=node, y=cause, covar=confounders, method=corr_type
    )["r"].values[0]


class CorrelationScorer(DecomposableScorer):
    """
    Score nodes by correlation
    """

    def score_node(
        self,
        graph: Graph,
        series: Dict[Node, Sequence[float]],
        node: Node,
        data: CaseData,
    ) -> Score:
        series_node = series[node]
        correlation, p_value = pearsonr(series_node, series[data.sli])

        score = Score(abs(correlation))
        score["pearson"] = correlation
        score["p-value"] = p_value
        return score


class PartialCorrelationScorer(DecomposableScorer):
    """
    Score nodes by partial correlation coefficient
    """

    def score_node(
        self,
        graph: Graph,
        series: Dict[Node, Sequence[float]],
        node: Node,
        data: CaseData,
    ) -> Score:
        data_frame = pd.DataFrame(series)
        correlation = partial_correlation(data.sli, node, graph, data_frame)
        score = Score(abs(correlation))
        score["partial-correlation"] = correlation
        return score



"""
This file is edited from `oracle/learn/prob_rf.py` in
    [the open-source code](https://github.com/perfd/perfd).

Reference:
[1] Silvery Fu, Saurabh Gupta, Radhika Mittal, and Sylvia Ratnasamy. On the Use of
    ML for Blackbox System Performance Prediction. NSDI 2021
"""
from collections import defaultdict
from itertools import chain
from typing import List

import numpy as np
from sklearn.ensemble import RandomForestRegressor

from .base import GMM
from .base import GMMPredictor
from .base import SKLearnGMM


class ProbRF(GMMPredictor):
    """
    Regress with a random forest
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._forest = RandomForestRegressor(random_state=self._seed)
        self._leaf2samples = defaultdict(list)

    def train(self, train_x: np.ndarray, train_y: np.ndarray):
        """
        Cluster train_y based on the forest
        """
        self._forest.fit(train_x, train_y)
        self._leaf2samples = defaultdict(list)
        indexes = self._forest.apply(train_x)
        for sample, leaves in zip(train_y, indexes):
            for tree_id, leaf in enumerate(leaves):
                self._leaf2samples[(tree_id, leaf)].append(sample.reshape(1))

    def predict(self, test_x: np.ndarray) -> List[GMM]:
        """
        Walk through the leave nodes and return the raw samples
        """
        indexes = self._forest.apply(test_x)
        gmms: List[GMM] = []
        for leaves in indexes:
            samples = list(
                chain(
                    *(
                        self._leaf2samples[(tree_id, leaf)]
                        for tree_id, leaf in enumerate(leaves)
                    )
                )
            )
            gmms.append(SKLearnGMM(samples=samples, gmm_k=self._gmm_k, seed=self._seed))
        return gmms



"""
Structral graph
"""
import dataclasses
from itertools import chain
from itertools import product
import logging
from typing import Dict
from typing import List
from typing import Set

import networkx as nx

from .base import GraphFactory
from ..model.case import CaseData
from ..model.graph import Graph
from ..model.graph import MemoryGraph
from ..model.graph import Node
from ..utils import YamlComposeLoader
from ..utils import topological_sort


@dataclasses.dataclass
class MetaVariable:
    """
    MetaVariable includes the following information
    - type: The meta variable type
    - component: Name of the component
    """

    type: str = dataclasses.field(default_factory=str)
    component: str = dataclasses.field(default_factory=str)

    def __hash__(self) -> int:
        return hash((self.type, self.component))


@dataclasses.dataclass
class ComponentConfig:
    """
    ComponentConfig includes the following information
    - name: Name of the component
    - dependencies: A mapping from one caller to its callees, and
    - mappings: A mapping from one variable to meta variables
    """

    name: str = dataclasses.field(default_factory=str)
    dependencies: Dict[str, Set[str]] = dataclasses.field(default_factory=dict)
    mappings: Dict[str, List[MetaVariable]] = dataclasses.field(default_factory=dict)

    def __init__(self, component: dict, types: set):
        self.name = component["name"]
        dependencies: dict = component.get("dependencies", {})
        self.dependencies = {
            caller: set(callees) for caller, callees in dependencies.items()
        }
        mappings: dict = component.get("mappings", {})
        self.mappings = self.parse_mappings(mappings, types)

    @staticmethod
    def parse_mappings(
        mappings: Dict[str, List[dict]], types: set
    ) -> Dict[str, List[MetaVariable]]:
        """
        Check whether a meta variable is valid
        """
        invalid_meta_variables = []
        ret: dict = {}
        for variable, meta_variables in mappings.items():
            mapped_meta_variables = []
            for meta_variable in meta_variables:
                mv_type = meta_variable.get("type", None)
                if mv_type not in types or "component" not in meta_variable:
                    invalid_meta_variables.append(meta_variable)
                    continue
                mapped_meta_variables.append(
                    MetaVariable(type=mv_type, component=meta_variable["component"])
                )
            ret[variable] = mapped_meta_variables

        if invalid_meta_variables:
            logger = logging.getLogger(__name__)
            logger.warning(
                "There are unknown meta variables in the metric mapping: %s",
                invalid_meta_variables,
            )
        return ret


@dataclasses.dataclass
class Config:
    """
    Config includes the following information to construct the structural graph
    - causal assumptions,
    - the component call graph, and
    - the mapping between variables and meta variables
    """

    types: Set[str] = dataclasses.field(default_factory=set)

    assumed_graph: Dict[str, Set[str]] = dataclasses.field(default_factory=dict)
    assumed_parents: Dict[str, Set[str]] = dataclasses.field(default_factory=dict)
    assumed_children: Dict[str, Set[str]] = dataclasses.field(default_factory=dict)
    assumed_ancestors: Dict[str, Set[str]] = dataclasses.field(default_factory=dict)
    assumed_descendents: Dict[str, Set[str]] = dataclasses.field(default_factory=dict)

    components: List[ComponentConfig] = dataclasses.field(default_factory=list)

    def __init__(self, filename: str):
        config = YamlComposeLoader.load(filename)
        self.types = set(config.get("types", []))
        assumptions: dict = config.get("assumptions", {})
        for item in ["graph", "parents", "children", "ancestors", "descendents"]:
            setattr(
                self,
                f"assumed_{item}",
                self.parse_assumptions(assumptions.get(item, {}), self.types),
            )
        components: list = config.get("components", [])
        self.components = [
            ComponentConfig(component, self.types) for component in components
        ]

    @staticmethod
    def parse_assumptions(assumptions: dict, types: set) -> Dict[str, set]:
        """
        A set of assumptions is a mapping from one meta variable type to its effects
        """
        logger = logging.getLogger(__name__)
        extra_causes = []
        extra_effects = {}
        ret = {mv_type: set() for mv_type in types}
        for cause, effects in assumptions.items():
            if cause not in types:
                extra_causes.append(cause)
                continue
            effects = set(effects)
            extra = effects - types
            if extra:
                extra_effects[cause] = extra
            ret[cause] = effects & types

        if extra_causes:
            logger.warning(
                "There are causes with unknown meta variables types"
                " in the assumptions: %s",
                extra_causes,
            )
        if extra_effects:
            logger.warning(
                "There are effects with unknown meta variables types"
                " in the assumptions: %s",
                extra_effects,
            )
        return ret

    def call_graph(self) -> nx.DiGraph:
        """
        Summary the dependencies of all components as the call graph
        """
        graph = nx.DiGraph()
        for component in self.components:
            graph.add_node(component.name)
            for caller, callees in component.dependencies.items():
                graph.add_edges_from(product([caller], callees))
        return graph


class StructuralGraph:
    """
    Structural graph derived from architecture
    """

    def __init__(self, filename: str = None, config: Config = None):
        """
        config: The necessary information to construct the structural graph
        filename: Mandatory if config is missing
        """
        if config is None:
            config = Config(filename)
        self._config = config
        self._skeleton = self.create_skeleton(config)

    @staticmethod
    def create_skeleton(config: Config) -> nx.DiGraph:
        """
        Create skeleton among meta variables
        """
        # 1. Collect the call graph
        call_graph = config.call_graph()

        skeleton = nx.DiGraph()
        # 2. Instantiate meta variables
        for mv_type in config.types:
            for component in call_graph.nodes:
                skeleton.add_node(MetaVariable(type=mv_type, component=component))

        # 3. Add edges based on causal assumptions
        for callee in call_graph.nodes:
            # 3.1 Within a component
            for cause_mv_type, effect_mv_types in config.assumed_graph.items():
                for effect_mv_type in effect_mv_types:
                    skeleton.add_edge(
                        MetaVariable(type=cause_mv_type, component=callee),
                        MetaVariable(type=effect_mv_type, component=callee),
                    )
            for callee_mv_type in config.types:
                # 3.2 For one-hop relations
                for caller in call_graph.predecessors(callee):
                    for caller_mv_type in config.assumed_parents[callee_mv_type]:
                        skeleton.add_edge(
                            MetaVariable(type=caller_mv_type, component=caller),
                            MetaVariable(type=callee_mv_type, component=callee),
                        )
                    for caller_mv_type in config.assumed_children[callee_mv_type]:
                        skeleton.add_edge(
                            MetaVariable(type=callee_mv_type, component=callee),
                            MetaVariable(type=caller_mv_type, component=caller),
                        )
                # 3.3 For multi-hop relations
                for caller in nx.ancestors(call_graph, callee):
                    for caller_mv_type in config.assumed_ancestors[callee_mv_type]:
                        skeleton.add_edge(
                            MetaVariable(type=caller_mv_type, component=caller),
                            MetaVariable(type=callee_mv_type, component=callee),
                        )
                    for caller_mv_type in config.assumed_descendents[callee_mv_type]:
                        skeleton.add_edge(
                            MetaVariable(type=callee_mv_type, component=callee),
                            MetaVariable(type=caller_mv_type, component=caller),
                        )

        return skeleton

    def _map_variable_meta(self, mask: Dict[str, Set[str]] = None):
        if mask is None:
            mask = {}

        variable2meta: Dict[Node, List[MetaVariable]] = {}
        meta2variable: Dict[MetaVariable, List[Node]] = {
            meta_variable: [] for meta_variable in self._skeleton.nodes
        }
        for component in self._config.components:
            for variable_name, meta_variables in component.mappings.items():
                variable = Node(entity=component.name, metric=variable_name)
                if component.name not in mask or variable_name in mask[component.name]:
                    variable2meta[variable] = meta_variables
                    for meta_variable in meta_variables:
                        meta2variable[meta_variable].append(variable)
        return variable2meta, meta2variable

    def visit(self, mask: Dict[str, Set[str]] = None) -> nx.DiGraph:
        """
        Generate the graph among the (component, variable) tuples
        """
        # 1. Set up mappings between variables and meta variables with the mask
        variable2meta, meta2variable = self._map_variable_meta(mask)

        graph = nx.DiGraph()
        visible_variables: Dict[MetaVariable, Set[Node]] = {
            meta_variable: set() for meta_variable in meta2variable
        }
        counter = {
            variable: len(meta_variables)
            for variable, meta_variables in variable2meta.items()
        }
        variable2visited: Dict[Node, Set[MetaVariable]] = {
            variable: set() for variable in counter
        }
        # 2. Iterate over meta variables in the topological order
        meta_variables: List[Set[MetaVariable]] = topological_sort(
            nodes=self._skeleton.nodes,
            predecessors=self._skeleton.predecessors,
            successors=self._skeleton.successors,
        )
        for meta_variable in chain(*meta_variables):
            current: Set[Node] = set()
            parents: Set[Node] = set()

            # 2.1 Handle multi-mapping variables
            for variable in meta2variable[meta_variable]:
                if counter[variable] == 1:
                    current.add(variable)
                elif len(variable2visited[variable]) == counter[variable] - 1:
                    # This is the last time to vist this variable
                    parents.add(variable)
                    for visited_mv in variable2visited[variable]:
                        graph.add_edges_from(
                            product(visible_variables[visited_mv], [variable])
                        )
                else:
                    # Skip to avoid self-loop
                    variable2visited[variable].add(meta_variable)

            # 2.2 Collect parents
            for cause_meta_variable in self._skeleton.predecessors(meta_variable):
                parents |= visible_variables[cause_meta_variable]

            # 2.3 Link from parents to current
            graph.add_edges_from(product(parents, current))

            # 2.4 Set visible variables for the following meta variables
            visible_variables[meta_variable] = current if current else parents

        return graph


class StructuralGraphFactory(GraphFactory):
    """
    Create Graph instances based on StructuralGraph
    """

    def __init__(self, structural_graph: StructuralGraph, **kwargs):
        super().__init__(**kwargs)
        self._structural_graph = structural_graph

    def create(self, data: CaseData, current: float) -> Graph:
        graph = self._structural_graph.visit(
            {
                entity: set(metric_names)
                for entity, metric_names in data.data_loader.metrics.items()
            }
        )
        return MemoryGraph(graph)



"""
Causal Inference-based Root Cause Analysis (CIRCA)
"""
from typing import Callable
from typing import Dict
from typing import Sequence
from typing import Tuple

import numpy as np

from .anm import ANMRegressor
from .base import Regressor
from ..base import Score
from ..base import Scorer
from ..common import DecomposableScorer
from ..common import zscore_conf
from ...model.case import CaseData
from ...model.graph import Graph
from ...model.graph import Node


class RHTScorer(DecomposableScorer):
    """
    Scorer with regression-based hypothesis testing
    """

    def __init__(
        self,
        tau_max: int = 0,
        regressor: Regressor = None,
        use_confidence: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._tau_max = max(tau_max, 0)
        self._regressor = regressor if regressor else ANMRegressor()
        self._use_confidence = use_confidence

    @staticmethod
    def _split_train_test(
        series_x: np.ndarray,
        series_y: np.ndarray,
        train_window: int,
        test_window: int,
    ):
        train_x: np.ndarray = series_x[:train_window, :]
        train_y: np.ndarray = series_y[:train_window]
        test_x: np.ndarray = series_x[-test_window:, :]
        test_y: np.ndarray = series_y[-test_window:]
        return train_x, test_x, train_y, test_y

    def split_data(
        self,
        data: Dict[Node, Sequence[float]],
        node: Node,
        parents: Sequence[Node],
        case_data: CaseData,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Split data for training and testing

        Return (train_x, test_x, train_y, test_y)
        """
        series = np.array([data[parent] for parent in parents if parent in data]).T
        if len(series) == 0:
            series = np.zeros((0, 0))
        series_x = np.hstack([np.roll(series, i, 0) for i in range(self._tau_max + 1)])
        series_x: np.ndarray = series_x[self._tau_max :, :]
        series_y = np.array(data[node][self._tau_max :])

        return self._split_train_test(
            series_x=series_x,
            series_y=series_y,
            train_window=case_data.train_window - self._tau_max,
            test_window=case_data.test_window,
        )

    def score_node(
        self,
        graph: Graph,
        series: Dict[Node, Sequence[float]],
        node: Node,
        data: CaseData,
    ) -> Score:
        parents = list(graph.parents(node))

        train_x, test_x, train_y, test_y = self.split_data(series, node, parents, data)
        z_scores = self._regressor.score(
            train_x=train_x, test_x=test_x, train_y=train_y, test_y=test_y
        )
        z_score = self._aggregator(abs(z_scores))
        confidence = zscore_conf(z_score)
        if self._use_confidence:
            score = Score(confidence)
            score.key = (score.score, z_score)
        else:
            score = Score(z_score)
        score["z-score"] = z_score
        score["Confidence"] = confidence

        return score


class DAScorer(Scorer):
    """
    Scorer with descendant adjustment
    """

    def __init__(
        self,
        threshold: float = 0,
        aggregator: Callable[[Sequence[float]], float] = max,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._threshold = max(threshold, 0.0)
        self._aggregator = aggregator

    def score(
        self,
        graph: Graph,
        data: CaseData,
        current: float,
        scores: Dict[Node, Score] = None,
    ) -> Dict[Node, Score]:
        sorted_nodes = [
            {node for node in nodes if node in scores}
            for nodes in graph.topological_sort
        ]
        # 0. Set topological rank
        for index, nodes in enumerate(sorted_nodes):
            for node in nodes:
                score = scores[node]
                score["index"] = index

        # 1. Gather child scores
        child_scores: Dict[Node, Dict[Node, float]] = {}
        for nodes in reversed(sorted_nodes):
            for node in nodes:
                child_score: Dict[Node, float] = {}
                for child in graph.children(node):
                    if child in scores:
                        child_score[child] = scores[child].score
                        if scores[child].score < self._threshold:
                            child_score.update(child_scores.get(child, {}))
                child_scores[node] = child_score

        # 2. Set child_score
        for node, score in scores.items():
            if score.score >= self._threshold:
                child_score = child_scores[node]
                if child_score:
                    child_score = self._aggregator(child_score.values())
                    score.score += child_score
                    score["child_score"] = child_score

        # 3. Set key
        for score in scores.values():
            score.key = (score.score, -score["index"], score.get("z-score", 0))
        return scores



"""
Compare algorithm combinations
"""
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import as_completed
import logging
from multiprocessing import Process
import os
from typing import Dict
from typing import List
from typing import Set
from typing import Tuple

from ...alg.common import Evaluation
from ...alg.common import Model
from ...alg.common import evaluate
from ...graph import GraphFactory
from ...model.case import Case
from ...utils import Timer
from ...utils import dump_csv
from ...utils import dump_json
from ...utils import load_csv
from ...utils import require_logging


IDEAL_METHOD = "GT-Ideal"


def _create_graph(
    graph_factory: GraphFactory,
    case: Case,
    timer_name: str,
    graph_filename: str,
    current: float,
):
    with Timer(name=timer_name):
        graph = graph_factory.create(data=case.data, current=current)
    graph.dump(graph_filename)


def _create_graphs(
    graph_factories: Dict[str, GraphFactory],
    cases: List[Case],
    delay: int,
    output_dir: str,
    timeout: int = 3600,
):
    logger = logging.getLogger(f"{run.__module__}.create_graphs")
    for graph_name, graph_factory in graph_factories.items():
        for index, case in enumerate(cases):
            case_output_dir = os.path.join(output_dir, str(index))
            graph_filename = os.path.join(case_output_dir, f"{graph_name}.json")
            if os.path.isfile(graph_filename):
                continue
            os.makedirs(case_output_dir, exist_ok=True)

            task = Process(
                target=require_logging(_create_graph),
                kwargs=dict(
                    graph_factory=graph_factory,
                    case=case,
                    timer_name=f"{graph_name} for case {index}",
                    graph_filename=graph_filename,
                    current=case.data.detect_time + delay,
                ),
            )
            task.start()
            task.join(timeout=timeout)
            if task.is_alive():
                task.terminate()
                logger.warning("%s: Timeout for case %d", graph_name, index)
                dump_json(graph_filename, {"status": "Timeout"})


def report_evaluation(name: str, report: Evaluation, duration: float = None):
    """
    Return a tuple of (name, AC@1, AC@3, AC@5, Avg@5).
    Duration will show at the end of the tuple if provided.
    """
    score = (
        name,
        report.accuracy(1),
        report.accuracy(3),
        report.accuracy(5),
        report.average(5),
    )
    if duration is not None:
        score += (duration,)
    return score


def report_ideal(cases: List[Case], name: str = IDEAL_METHOD):
    """
    Return the score tuple for the ideal algorithm
    """
    report = Evaluation()
    for case in cases:
        report(ranks=list(case.answer), answers=case.answer)
    return report_evaluation(name, report)


def _wrap_ideal_report(scores: List[tuple], cases: List[Case]):
    names = {score[0] for score in scores}
    if IDEAL_METHOD not in names:
        scores = scores + [report_ideal(cases=cases, name=IDEAL_METHOD)]
    return scores


def _evaluate(models: List[Model], cases: List[Case], **kwargs):
    logger = logging.getLogger(f"{run.__module__}.evaluate")
    scores: List[Tuple[str, float, float, float, float, float]] = []
    num_cases = max(len(cases), 1)
    for model in models:
        name = model.name
        with Timer(name=name) as timer:
            report = evaluate(model, cases, **kwargs)
            duration = timer.duration
        score = report_evaluation(name, report, duration.total_seconds() / num_cases)
        logger.info("Finish: %s,%f,%f,%f,%f,%f", *score)
        scores.append(score)
    return scores


def _load_cached_report(report_filename: str, models: List[Model]):
    scores: List[Tuple[str, float, float, float, float, float]] = []
    cached_models: Set[str] = set()
    if os.path.exists(report_filename):
        reader = load_csv(report_filename)
        _ = next(reader)
        for model_name, *values in reader:
            scores.append((model_name, *map(float, values)))
            cached_models.add(model_name)
    models = [model for model in models if model.name not in cached_models]
    return scores, models


def run(
    models: List[Model],
    cases: List[Case],
    graph_factories: Dict[str, GraphFactory] = None,
    output_dir: str = None,
    report_filename: str = "report.csv",
    delay: int = 300,
    max_workers: int = 1,
    timeout: int = 3600,
):
    # pylint: disable=too-many-arguments
    """
    Compare different models

    Parameters:
        graph_factories: If given, graphs will be created for each case
            before any model starts.
        output_dir: Where the intermediate results will be cached.
        report_filename: The name of a csv file that will store the experiment results.
        delay: A model is assumed to start in this amount of seconds
            after a fault is detected.
        max_workers: The given number of processes will be created
            to accelerate the calculation.
    """
    params = dict(delay=delay, output_dir=output_dir, timeout=timeout)
    if output_dir and graph_factories is not None:
        if max_workers >= 2:
            graph_factory_items = list(graph_factories.items())
            with ProcessPoolExecutor(max_workers=max_workers) as executor:
                tasks = [
                    executor.submit(
                        require_logging(_create_graphs),
                        graph_factories=dict(graph_factory_items[i::max_workers]),
                        cases=cases,
                        **params,
                    )
                    for i in range(max_workers)
                ]
                for task in as_completed(tasks):
                    task.result()
        else:
            _create_graphs(
                graph_factories=graph_factories,
                cases=cases,
                **params,
            )

    scores, models = _load_cached_report(report_filename=report_filename, models=models)
    if max_workers >= 2:
        with ProcessPoolExecutor(max_workers=max_workers) as executor:
            tasks = [
                executor.submit(
                    require_logging(_evaluate),
                    models=models[i::max_workers],
                    cases=cases,
                    **params,
                )
                for i in range(max_workers)
            ]
            for task in as_completed(tasks):
                scores += task.result()
    else:
        scores += _evaluate(models=models, cases=cases, **params)
    scores = _wrap_ideal_report(scores=scores, cases=cases)

    dump_csv(
        filename=report_filename,
        data=scores,
        headers=["method", "AC@1", "AC@3", "AC@5", "Avg@5", "avg duration"],
    )



"""
Simluation with vector auto-regression model
"""
import datetime
import os
import pickle
from typing import Callable
from typing import List
from typing import Set

import networkx as nx
import numpy as np
from sklearn.preprocessing import StandardScaler

from ...graph.common import StaticGraphFactory
from ...model.case import Case
from ...model.case import CaseData
from ...model.data_loader import MemoryDataLoader
from ...model.graph import MemoryGraph
from ...model.graph import Node
from ...utils import ENCODING
from ...utils import dump_csv
from ...utils import dump_json
from ...utils import load_csv
from ...utils import load_json


_SLI = 0
ENTITY = "SIM"


class SimCase(Case):
    """
    Simulated case data for evaluation
    """

    DATA_FILENAME = "data.csv"
    INFO_FILENAME = "info.json"
    DETAIL_FILENAME = "details.pk"

    def __init__(
        self,
        data: np.ndarray,
        causes: Set[int],
        length_normal: int,
        interval: datetime.timedelta = datetime.timedelta(minutes=1),
        case_data_params: dict = None,
        details: dict = None,
    ):
        # pylint: disable=too-many-arguments
        super().__init__(
            data=None,
            answer={Node(entity=ENTITY, metric=str(cause)) for cause in causes},
        )
        self._time_series = data
        self._causes = causes
        self._length_normal = length_normal
        self._interval = interval
        self._case_data_params = {} if case_data_params is None else case_data_params
        self._details = details

    @property
    def data(self) -> CaseData:
        if self._data is not None:
            return self._data

        _, num_node = self._time_series.shape
        interval = self._interval.total_seconds()
        data = {
            ENTITY: {
                str(node): [
                    (i * interval, value)
                    for i, value in enumerate(self._time_series[:, node])
                ]
                for node in range(num_node)
            }
        }
        self._data = CaseData(
            data_loader=MemoryDataLoader(data=data),
            sli=Node(entity=ENTITY, metric=str(_SLI)),
            detect_time=self._length_normal * interval,
            interval=self._interval,
            **self._case_data_params,
        )
        return self._data

    @classmethod
    def load(cls, folder: str, **kwargs) -> "SimCase":
        """
        Load from a folder

        Parameters:
            folder: where the case is dumped
            **kwargs: Other parameters will be passed to construct SimCase
        """
        filename = os.path.join(folder, cls.DATA_FILENAME)
        data = np.array(list(load_csv(filename))).astype(float)
        info: dict = load_json(os.path.join(folder, cls.INFO_FILENAME))
        return SimCase(
            data=data,
            causes=info["causes"],
            length_normal=info["length_normal"],
            **kwargs,
        )

    def dump(self, folder: str):
        """
        Dump into a folder
        """
        # Dump data into csv
        os.makedirs(folder, exist_ok=True)
        dump_csv(
            filename=os.path.join(folder, self.DATA_FILENAME), data=self._time_series
        )
        dump_json(
            filename=os.path.join(folder, self.INFO_FILENAME),
            data=dict(causes=list(self._causes), length_normal=self._length_normal),
        )
        if self._details:
            with open(os.path.join(folder, self.DETAIL_FILENAME), "wb") as obj:
                pickle.dump(self._details, obj)


class SimDataset:
    """
    A combination of graph and simulated cases
    """

    _INDEX_FILENAME = "index"
    GRAPH_FILENAME = "graph.json"
    CASES_FOLDER = "cases"

    def __init__(self, graph: MemoryGraph, cases: List[SimCase]):
        self._cases = cases
        self._graph = graph
        self._graph_factory = StaticGraphFactory(graph=graph)

    @property
    def cases(self) -> List[Case]:
        """
        Simulated cases
        """
        return self._cases

    @property
    def graph_factory(self) -> StaticGraphFactory:
        """
        The causal graph in the data generation process
        """
        return self._graph_factory

    @classmethod
    def load(cls, folder: str, **kwargs) -> "SimDataset":
        """
        Load from a folder

        Parameters:
            folder: where the case is dumped
            **kwargs: Other parameters will be passed to construct SimCase
        """
        cases_folder = os.path.join(folder, cls.CASES_FOLDER)
        graph = MemoryGraph.load(os.path.join(folder, cls.GRAPH_FILENAME))
        with open(
            os.path.join(cases_folder, cls._INDEX_FILENAME), encoding=ENCODING
        ) as obj:
            num_cases = int(next(obj))
        cases = [
            SimCase.load(os.path.join(cases_folder, str(index)), **kwargs)
            for index in range(num_cases)
        ]
        return SimDataset(graph=graph, cases=cases)

    def dump(self, folder: str):
        """
        Dump into a folder
        """
        # Dump data into csv
        cases_folder = os.path.join(folder, self.CASES_FOLDER)
        os.makedirs(cases_folder, exist_ok=True)
        self._graph.dump(os.path.join(folder, self.GRAPH_FILENAME))

        with open(
            os.path.join(cases_folder, self._INDEX_FILENAME), "w", encoding=ENCODING
        ) as obj:
            obj.write(str(len(self._cases)))
        for index, case in enumerate(self._cases):
            case.dump(os.path.join(cases_folder, str(index)))


def _normal_weight(rng: np.random.Generator) -> float:
    weight = rng.standard_normal()
    return np.sign(weight) * (abs(weight) + 0.2)


def _uniform_weight(rng: np.random.Generator) -> float:
    segments = [(-2.0, -0.5), (0.5, 2.0)]
    low, high = rng.choice(segments)
    return rng.uniform(low=low, high=high)


def generate_sedag(
    num_node: int,
    num_edge: int,
    weight_generator: Callable[[np.random.Generator], float] = _uniform_weight,
    rng: np.random.Generator = None,
) -> np.ndarray:
    """
    Generate a weighted directed acyclic graph with a single end.

    The first node with index of 0 is the only end that does not have results.

    Returns:
        a matrix, where matrix[i, j] != 0 means j is the cause of i
    """
    num_edge = min(max(num_edge, num_node - 1), int(num_node * (num_node - 1) / 2))
    if rng is None:
        rng = np.random.default_rng()
    matrix = np.zeros((num_node, num_node))
    # Make the graph connected
    for cause in range(1, num_node):
        result = rng.integers(low=0, high=cause)
        matrix[result, cause] = weight_generator(rng)
    num_edge -= num_node - 1
    while num_edge > 0:
        cause = rng.integers(low=1, high=num_node)
        result = rng.integers(low=0, high=cause)
        if not matrix[result, cause]:
            matrix[result, cause] = weight_generator(rng)
            num_edge -= 1
    return matrix


def generate_case(
    weight: np.ndarray,
    length_normal: int = 1440,
    fault_duration: int = 2,
    length_abnormal: int = 10,
    beta: float = 1e-1,
    tau: float = 3,
    sigmas: np.ndarray = None,
    fault: np.ndarray = None,
    rng: np.random.Generator = None,
) -> SimCase:
    # pylint: disable=too-many-arguments, too-many-locals
    """
    Generate a case

    Parameters:
        weight: The inversed matrix of I - A
    """
    if rng is None:
        rng = np.random.default_rng()
    length_abnormal = max(length_abnormal, fault_duration)

    num_node, _ = weight.shape
    data: np.ndarray = np.zeros((0, num_node))
    if sigmas is None:
        sigmas = rng.standard_exponential(num_node)

    values = rng.standard_normal(num_node) * sigmas

    # Generate a series of x with x^{(t)} = A x^{(t)} + x^{(t - 1)} + epsilon^{(t)}
    # or x^{(t)} = A^{(\prime)} (x^{(t - 1)} + epsilon^{(t)})
    # where A^{(\prime)} = \sum_{i=0}^{num_node} A^{i}
    for _ in range(length_normal):
        values = weight @ (beta * values + rng.standard_normal(num_node) * sigmas)
        data = np.append(data, [values], axis=0)

    sli_mean: float = data[:, _SLI].mean()
    sli_sigma: float = data[:, _SLI].std()
    # Inject a fault
    if fault is None:
        num_causes = min(rng.poisson(1) + 1, num_node)
        causes = rng.choice(num_node, size=num_causes, replace=False)
        fault = np.zeros(num_node)
        alpha = rng.standard_exponential(size=num_causes)
        epsilon = rng.standard_normal(num_node)
        while True:
            fault[causes] = alpha
            sli_value: float = np.dot(
                weight[_SLI, :], beta * values + (epsilon + fault) * sigmas
            )
            if abs(sli_value - sli_mean) > tau * sli_sigma:
                break
            alpha *= 2
    else:
        causes: np.ndarray = np.where(fault)[0]
        assert causes.size

    # Faulty data
    for _ in range(fault_duration):
        values = weight @ (
            beta * values + (rng.standard_normal(num_node) + fault) * sigmas
        )
        data = np.append(data, [values], axis=0)
    for _ in range(length_abnormal - fault_duration):
        values = weight @ (beta * values + rng.standard_normal(num_node) * sigmas)
        data = np.append(data, [values], axis=0)

    scaler = StandardScaler().fit(data[:length_normal, :])
    data = np.around(scaler.transform(data), decimals=3)

    details = dict(
        fault=fault,
        sigmas=sigmas,
        stds=scaler.scale_,
        weight=weight,
    )
    return SimCase(
        data=data,
        causes=set(causes.tolist()),
        length_normal=length_normal,
        details=details,
    )


def generate(
    num_node: int, num_edge: int, num_cases: int = 100, rng: np.random.Generator = None
) -> SimDataset:
    """
    Generate a dataset with the same graph and serveral cases
    """
    if rng is None:
        rng = np.random.default_rng()

    # A = Generate weighted DAG
    matrix = generate_sedag(num_node=num_node, num_edge=num_edge, rng=rng)
    prod = np.eye(num_node)
    weight = np.eye(num_node)  # The reversed matrix of I - A
    for _ in range(1, num_node):
        prod = prod @ matrix
        weight += prod

    cases = [generate_case(weight=weight, rng=rng) for _ in range(num_cases)]
    graph = nx.DiGraph(
        (
            (
                Node(entity=ENTITY, metric=str(cause)),
                Node(entity=ENTITY, metric=str(result)),
            )
            for result, cause in zip(*np.where(matrix))
        )
    )
    return SimDataset(cases=cases, graph=MemoryGraph(graph))



"""
Utilities
"""
from abc import ABC
from itertools import product
from typing import Dict
from typing import Generator
from typing import List
from typing import Tuple

from sklearn.svm import SVR

from . import utils
from ...alg.base import Scorer
from ...alg.common import Model
from ...alg.common import NSigmaScorer
from ...alg.correlation import CorrelationScorer
from ...alg.correlation import PartialCorrelationScorer
from ...alg.invariant_network import CRDScorer
from ...alg.invariant_network import ENMFScorer
from ...alg.dfs import DFSScorer
from ...alg.dfs import MicroHECLScorer
from ...alg.evt import SPOTScorer
from ...alg.random_walk import RandomWalkScorer
from ...alg.random_walk import SecondOrderRandomWalkScorer
from ...alg.ci import DAScorer
from ...alg.ci import RHTScorer
from ...alg.ci.anm import ANMRegressor
from ...alg.ci.gmm import GMMRegressor
from ...alg.ci.gmm.mdn import MDNPredictor
from ...graph import GraphFactory
from ...graph.common import EmptyGraphFactory
from ...graph.pcts import PCTSFactory
from ...graph.r import PCAlgFactory
from ...graph.structural import StructuralGraphFactory


EMPTY_GRAPH_NAME = "Empty"


class ModelGetter(ABC):
    """
    Abstract interface to get models
    """

    def __init__(self, params: utils.Params):
        self._params = params

    @staticmethod
    def compose_parameters(
        values: tuple, names: tuple, abbrs: tuple
    ) -> Tuple[str, dict]:
        """
        Wrap parameters as (suffix str and parameter dict)
        """
        suffix = ""
        params = {}
        for value, name, abbr in zip(values, names, abbrs):
            if value is None:
                continue
            params[name] = value
            if isinstance(value, bool):
                if value:
                    suffix += f"_{abbr}"
            else:
                suffix += f"_{abbr}{value}"
        return (suffix, params)

    @staticmethod
    def compose_fields(
        fields: Tuple[tuple], names: tuple, abbrs: tuple
    ) -> Generator[Tuple[str, dict], None, None]:
        """
        Generate parameter combinations
        """
        for values in product(*fields):
            yield ModelGetter.compose_parameters(values, names=names, abbrs=abbrs)

    def _get(
        self,
        graph_factories: Dict[str, GraphFactory] = None,
        **scorer_params,
    ) -> List[Model]:
        raise NotImplementedError

    def get(
        self,
        graph_factory_params: dict,
        graph_factories: Dict[str, GraphFactory] = None,
        **scorer_params,
    ) -> Tuple[List[Model], Dict[str, GraphFactory]]:
        # pylint: disable=unused-argument
        """
        List models and used graph factories
        """
        if self._params is None:
            return [], {}
        return self._get(graph_factories=None, **scorer_params), {}


def _get_graph_factories(
    graph_params: utils.GraphParams, seed: int, structural_graph_params: dict = None
):
    graph_factories: Dict[str, GraphFactory] = {}

    if graph_params.pc_gauss:
        for suffix, params in ModelGetter.compose_fields(
            (graph_params.pc_gauss.alpha, graph_params.pc_gauss.max_conds_dim),
            names=("alpha", "max_conds_dim"),
            abbrs=("a", "m"),
        ):
            graph_factories["PC_gauss" + suffix] = PCAlgFactory(
                method="PC-gauss",
                num_cores=graph_params.pc_gauss.num_cores,
                seed=seed,
                **params,
            )
    if graph_params.pc_gsq:
        for suffix, params in ModelGetter.compose_fields(
            (graph_params.pc_gsq.alpha, graph_params.pc_gsq.max_conds_dim),
            names=("alpha", "max_conds_dim"),
            abbrs=("a", "m"),
        ):
            graph_factories["PC_gsq" + suffix] = PCAlgFactory(
                method="PC-gsq",
                num_cores=graph_params.pc_gsq.num_cores,
                seed=seed,
                **params,
            )
    if graph_params.pcts:
        for suffix, params in ModelGetter.compose_fields(
            (
                graph_params.pcts.alpha,
                graph_params.pcts.max_conds_dim,
                graph_params.pcts.tau_max,
            ),
            names=("alpha", "max_conds_dim", "tau_max"),
            abbrs=("a", "m", "t"),
        ):
            graph_factories["PCTS" + suffix] = PCTSFactory(seed=seed, **params)
    if graph_params.structural and structural_graph_params is not None:
        graph_factories["Structural"] = StructuralGraphFactory(
            **structural_graph_params, seed=seed
        )

    return graph_factories


def _get_detectors(ad_params: utils.ADParams, **scorer_params):
    """
    Map a detector name to a pair of Scorer and threshold
    """
    detectors: Dict[str, Tuple[Scorer, float]] = {}

    if ad_params.nsigma:
        detectors["NSigma"] = (NSigmaScorer(**scorer_params), 3)
    if ad_params.spot:
        for suffix, params in ModelGetter.compose_fields(
            (ad_params.spot.risk,), names=("proba",), abbrs=("p",)
        ):
            detectors["SPOT" + suffix] = (SPOTScorer(**params, **scorer_params), 0)

    return detectors


class NSigmaGetter(ModelGetter):
    """
    Get NSigma models
    """

    def __init__(self, params: utils.NSigmaParams):
        super().__init__(params=params)
        self._params = params

    def _get(
        self,
        graph_factories: Dict[str, GraphFactory] = None,
        **scorer_params,
    ) -> List[Model]:
        graph_factory = EmptyGraphFactory()
        return [
            Model(
                graph_factory=graph_factory,
                scorers=[NSigmaScorer(**scorer_params)],
                names=(EMPTY_GRAPH_NAME, "NSigma"),
            )
        ]


class SPOTGetter(ModelGetter):
    """
    Get SPOT models
    """

    def __init__(self, params: utils.SPOTParams):
        super().__init__(params=params)
        self._params = params

    def _get(
        self,
        graph_factories: Dict[str, GraphFactory] = None,
        **scorer_params,
    ) -> List[Model]:
        graph_factory = EmptyGraphFactory()
        return [
            Model(
                graph_factory=graph_factory,
                scorers=[SPOTScorer(**params, **scorer_params)],
                names=(EMPTY_GRAPH_NAME, "SPOT" + suffix),
            )
            for suffix, params in ModelGetter.compose_fields(
                (self._params.risk,), names=("proba",), abbrs=("p",)
            )
        ]


class ScorerModelGetter(ModelGetter):
    """
    Get graph-based models
    """

    def __init__(self, params: utils.ScorerParams):
        super().__init__(params=params)
        self._params = params

    def _compose_parameters(self, **scorer_params):
        raise NotImplementedError

    def _get_model_base(
        self,
        graph_factories: Dict[str, GraphFactory],
        **scorer_params,
    ):
        for suffix, params in self._compose_parameters(**scorer_params):
            for graph_name, graph_factory in graph_factories.items():
                yield graph_factory, graph_name, suffix, params

    def _get(
        self,
        graph_factories: Dict[str, GraphFactory] = None,
        **scorer_params,
    ) -> List[Model]:
        raise NotImplementedError

    def get(
        self,
        graph_factory_params: dict,
        graph_factories: Dict[str, GraphFactory] = None,
        **scorer_params,
    ) -> Tuple[List[Model], Dict[str, GraphFactory]]:
        if self._params is None:
            return [], {}
        if graph_factories is None:
            graph_factories = _get_graph_factories(
                graph_params=self._params.graph, **graph_factory_params
            )
        return (
            self._get(graph_factories=graph_factories, **scorer_params),
            graph_factories,
        )


class DFSGetter(ScorerModelGetter):
    """
    Get DFS models
    """

    def __init__(self, params: utils.DFSParams):
        super().__init__(params=params)
        self._params = params

    def _compose_parameters(self, **scorer_params):
        detectors = _get_detectors(ad_params=self._params.detector, **scorer_params)
        for detector_name, (detector, anomaly_threshold) in detectors.items():
            suffix, params = self.compose_parameters(
                values=(anomaly_threshold,), names=("anomaly_threshold",), abbrs=("a",)
            )
            params.update(scorer_params)
            yield detector, detector_name, suffix, params

    def _get_model_base(
        self,
        graph_factories: Dict[str, GraphFactory],
        **scorer_params,
    ):
        for detector, detector_name, suffix, params in self._compose_parameters(
            **scorer_params
        ):
            for graph_name, graph_factory in graph_factories.items():
                yield (
                    graph_factory,
                    detector,
                    [graph_name, detector_name],
                    suffix,
                    params,
                )

    def _get(
        self,
        graph_factories: Dict[str, GraphFactory] = None,
        **scorer_params,
    ) -> List[Model]:
        return [
            Model(
                graph_factory=graph_factory,
                scorers=[detector, DFSScorer(**params)],
                names=[*names, "DFS" + suffix],
            )
            for graph_factory, detector, names, suffix, params in self._get_model_base(
                graph_factories=graph_factories, **scorer_params
            )
        ]


class MicroscopeGetter(DFSGetter):
    """
    Get Microscope models
    """

    def _get(
        self,
        graph_factories: Dict[str, GraphFactory] = None,
        **scorer_params,
    ) -> List[Model]:
        return [
            Model(
                graph_factory=graph_factory,
                scorers=[
                    detector,
                    DFSScorer(**params),
                    CorrelationScorer(**scorer_params),
                ],
                names=[*names, "DFS" + suffix, "Pearson"],
            )
            for graph_factory, detector, names, suffix, params in self._get_model_base(
                graph_factories=graph_factories, **scorer_params
            )
        ]


class MicroHECLGetter(DFSGetter):
    """
    Get MicroHECL models
    """

    def __init__(self, params: utils.MicroHECLParams):
        super().__init__(params=params)
        self._params = params

    def _compose_parameters(self, **scorer_params):
        detectors = _get_detectors(ad_params=self._params.detector, **scorer_params)
        for detector_name, (detector, anomaly_threshold) in detectors.items():
            for stop_threshold in self._params.stop_threshold:
                suffix, params = self.compose_parameters(
                    values=(anomaly_threshold, stop_threshold),
                    names=("anomaly_threshold", "stop_threshold"),
                    abbrs=("a", "s"),
                )
                params.update(scorer_params)
                yield detector, detector_name, suffix, params

    def _get(
        self,
        graph_factories: Dict[str, GraphFactory] = None,
        **scorer_params,
    ) -> List[Model]:
        return [
            Model(
                graph_factory=graph_factory,
                scorers=[detector, MicroHECLScorer(**params)],
                names=[*names, "MicroHECL" + suffix],
            )
            for graph_factory, detector, names, suffix, params in self._get_model_base(
                graph_factories=graph_factories, **scorer_params
            )
        ]


class RWModelGetter(ScorerModelGetter):
    """
    Get random walk-based models
    """

    def __init__(self, params: utils.RandomWalkParams):
        super().__init__(params=params)
        self._params = params

    def _compose_parameters(self, **scorer_params):
        for suffix, params in self.compose_fields(
            (self._params.rho,), names=("rho",), abbrs=("r",)
        ):
            params.update(scorer_params)
            yield suffix, params


class MicroCauseGetter(RWModelGetter):
    """
    Get MicroCause models
    """

    def __init__(self, params: utils.MicroCauseParams):
        super().__init__(params=params)
        self._params = params

    def _get(
        self,
        graph_factories: Dict[str, GraphFactory] = None,
        **scorer_params,
    ) -> List[Model]:
        return [
            Model(
                graph_factory=graph_factory,
                scorers=[
                    PartialCorrelationScorer(**scorer_params),
                    RandomWalkScorer(**params),
                ],
                names=[graph_name, "PartialCorrelation", "RW" + suffix],
            )
            for graph_factory, graph_name, suffix, params in self._get_model_base(
                graph_factories=graph_factories, **scorer_params
            )
        ]


class CloudRangerGetter(RWModelGetter):
    """
    Get CloudRanger models
    """

    def __init__(self, params: utils.CloudRangerParams):
        super().__init__(params=params)
        self._params = params

    def _compose_parameters(self, **scorer_params):
        for suffix, params in self.compose_fields(
            (self._params.rho, self._params.beta),
            names=("rho", "beta"),
            abbrs=("r", "b"),
        ):
            params.update(scorer_params)
            yield suffix, params

    def _get(
        self,
        graph_factories: Dict[str, GraphFactory] = None,
        **scorer_params,
    ) -> List[Model]:
        return [
            Model(
                graph_factory=graph_factory,
                scorers=[
                    CorrelationScorer(**scorer_params),
                    SecondOrderRandomWalkScorer(**params),
                ],
                names=[
                    graph_name,
                    "Pearson",
                    "RW_2" + suffix,
                ],
            )
            for graph_factory, graph_name, suffix, params in self._get_model_base(
                graph_factories=graph_factories, **scorer_params
            )
        ]


class ENMFGetter(ModelGetter):
    """
    Get ENMF models
    """

    def __init__(self, params: utils.ENMFParams):
        super().__init__(params=params)
        self._params = params

    def _compose_parameters(self, **scorer_params):
        for suffix, params in self.compose_fields(
            (
                self._params.use_softmax,
                self._params.gamma,
                self._params.tau,
                self._params.discrete,
            ),
            names=("use_softmax", "gamma", "tau", "discrete"),
            abbrs=("soft", "c", "t", "d"),
        ):
            use_softmax: bool = params.pop("use_softmax")
            discrete: bool = params.pop("discrete")
            params = dict(
                model_params=params,
                use_softmax=use_softmax,
                discrete=discrete,
                **scorer_params,
            )
            yield suffix, params

    def _get(
        self,
        graph_factories: Dict[str, GraphFactory] = None,
        **scorer_params,
    ) -> List[Model]:
        return [
            Model(
                graph_factory=EmptyGraphFactory(),
                scorers=[ENMFScorer(**params)],
                names=(EMPTY_GRAPH_NAME, "ENMF" + suffix),
            )
            for suffix, params in self._compose_parameters(**scorer_params)
        ]


class CRDGetter(ModelGetter):
    """
    Get CRD models
    """

    def __init__(self, params: utils.CRDParams):
        super().__init__(params=params)
        self._params = params

    def _compose_parameters(self, **scorer_params):
        for suffix, params in self.compose_fields(
            (
                self._params.gamma,
                self._params.tau,
                self._params.discrete,
                self._params.num_cluster,
                self._params.alpha,
                self._params.beta,
                self._params.learning_rate,
            ),
            names=(
                "gamma",
                "tau",
                "discrete",
                "num_cluster",
                "alpha",
                "beta",
                "learning_rate",
            ),
            abbrs=("c", "t", "d", "nc", "a", "b", "lr"),
        ):
            discrete: bool = params.pop("discrete")
            params = dict(
                model_params=params,
                discrete=discrete,
                **scorer_params,
            )
            yield suffix, params

    def _get(
        self,
        graph_factories: Dict[str, GraphFactory] = None,
        **scorer_params,
    ) -> List[Model]:
        return [
            Model(
                graph_factory=EmptyGraphFactory(),
                scorers=[CRDScorer(**params)],
                names=(EMPTY_GRAPH_NAME, "CRD" + suffix),
            )
            for suffix, params in self._compose_parameters(**scorer_params)
        ]


class CIModelGetter(ScorerModelGetter):
    """
    Get causal inference-based models
    """

    def __init__(self, params: utils.CIParams):
        super().__init__(params=params)
        self._params = params

    def _compose_parameters(self, **scorer_params):
        for suffix, params in self.compose_fields(
            (self._params.tau_max,), names=("tau_max",), abbrs=("t",)
        ):
            params.update(scorer_params)
            for regressor in self._params.regressor:
                if regressor == "linear":
                    yield suffix, params
                elif regressor == "svr":
                    yield f"_svr{suffix}", dict(
                        regressor=ANMRegressor(regressor=SVR(kernel="sigmoid")),
                        **params,
                    )
                elif regressor == "rf":
                    yield f"_rf{suffix}", dict(regressor=GMMRegressor(), **params)
                elif regressor == "mdn":
                    yield f"_mdn{suffix}", dict(
                        regressor=GMMRegressor(regressor=MDNPredictor()), **params
                    )


class RHTGetter(CIModelGetter):
    """
    Get RHT models
    """

    def _get(
        self,
        graph_factories: Dict[str, GraphFactory] = None,
        **scorer_params,
    ) -> List[Model]:
        return [
            Model(
                graph_factory=graph_factory,
                scorers=[RHTScorer(**params)],
                names=(graph_name, "RHT" + suffix),
            )
            for graph_factory, graph_name, suffix, params in self._get_model_base(
                graph_factories=graph_factories, **scorer_params
            )
        ]


class RHTDAGetter(CIModelGetter):
    """
    Get RHT-DA models
    """

    def _get(
        self,
        graph_factories: Dict[str, GraphFactory] = None,
        **scorer_params,
    ) -> List[Model]:
        # The three-sigma rule of thumb, as use_confidence=False by default
        threshold = 3
        ranker = DAScorer(threshold=threshold)
        return [
            Model(
                graph_factory=graph_factory,
                scorers=[RHTScorer(**params), ranker],
                names=(graph_name, "RHT" + suffix, "DA"),
            )
            for graph_factory, graph_name, suffix, params in self._get_model_base(
                graph_factories=graph_factories, **scorer_params
            )
        ]


def get_models(
    structural_graph_params: dict = None,
    graph_factories: Dict[str, GraphFactory] = None,
    params: utils.ModelParams = None,
    seed: int = 0,
    **scorer_params,
) -> Tuple[List[Model], Dict[str, GraphFactory]]:
    """
    Prepare algorithm candidates

    Parameters:
        graph_factories: Specify GraphFactory for models
        params: Specify options for model parameters
    """
    if params is None:
        params = utils.ModelParams()

    graph_factory_params = dict(
        structural_graph_params=structural_graph_params, seed=seed
    )
    getter_params = dict(
        graph_factory_params=graph_factory_params,
        graph_factories=graph_factories,
        seed=seed,
        **scorer_params,
    )
    getters: List[ModelGetter] = [
        NSigmaGetter(params.nsigma),
        SPOTGetter(params.spot),
        DFSGetter(params.dfs),
        MicroscopeGetter(params.micro_scope),
        MicroHECLGetter(params.micro_hecl),
        MicroCauseGetter(params.micro_cause),
        CloudRangerGetter(params.cloud_ranger),
        ENMFGetter(params.enmf),
        CRDGetter(params.crd),
        RHTGetter(params.rht),
        RHTDAGetter(params.rht_da),
    ]

    models: List[Model] = []
    graph2cache: Dict[str, GraphFactory] = {}
    for getter in getters:
        submodels, sub_graph_factories = getter.get(**getter_params)
        models += submodels
        graph2cache.update(sub_graph_factories)

    return models, graph2cache



"""
Scorers with the ground truth graph in the simulation dataset
"""
from typing import Dict
from typing import Sequence
from typing import Tuple

import numpy as np

from ...alg.ci import RHTScorer
from ...model.case import CaseData
from ...model.graph import Node


class SimRHTScorer(RHTScorer):
    """
    Update RHTScorer with the ground truth parents in the simulation dataset
        Pa(V_{i}^{(t)}) = Pa^{(t)}(V_{i}^{t}) \\cup {V_{i}^{(t - 1)}}
    """

    def split_data(
        self,
        data: Dict[Node, Sequence[float]],
        node: Node,
        parents: Sequence[Node],
        case_data: CaseData,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Split data for training and testing

        Return (train_x, test_x, train_y, test_y)
        """
        series = [data[parent] for parent in parents if parent in data]
        # from [x1, x2, ..., xn] to [xn, x1, x2, ...]
        series.append(np.roll(data[node], 1))
        series = np.array(series).T
        series_x: np.ndarray = series[1:, :]
        series_y = np.array(data[node][1:])

        return self._split_train_test(
            series_x=series_x,
            series_y=series_y,
            train_window=case_data.train_window - 1,
            test_window=case_data.test_window,
        )



"""
Abstract interfaces for graph generation
"""
from abc import ABC
from typing import Union

from ..model.case import CaseData
from ..model.graph import Graph
from ..model.graph import MemoryGraph


class GraphFactory(ABC):
    """
    The abstract interface to create Graph
    """

    def __init__(self, seed: int = 0):
        self._seed = seed

    @staticmethod
    def load(filename: str) -> Union[Graph, None]:
        """
        Load a graph from the given file

        Returns:
        - A graph, if available
        - None, if dump/load is not supported
        - Raise LoadingInvalidGraphException if the file cannot be parsed
        """
        return MemoryGraph.load(filename)

    def create(self, data: CaseData, current: float) -> Graph:
        """
        Create the Graph
        """
        raise NotImplementedError



"""
Common utilities
"""
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import as_completed
import logging
from multiprocessing import Process
from multiprocessing import Queue
import os
import queue
from typing import Dict
from typing import List
from typing import Sequence
from typing import Set
from typing import Tuple

import numpy as np
from scipy.stats import norm
from sklearn.preprocessing import StandardScaler

from .base import Score
from .base import Scorer
from ..graph import GraphFactory
from ..model.case import Case
from ..model.case import CaseData
from ..model.graph import Graph
from ..model.graph import LoadingInvalidGraphException
from ..model.graph import Node
from ..utils import dump_json
from ..utils import load_json
from ..utils import require_logging


def pearson(series_a: np.ndarray, series_b: np.ndarray) -> float:
    """
    Pearson coefficient, checking constant
    """
    std_a: float = series_a.std()
    std_b: float = series_b.std()
    if std_a == 0 or std_b == 0:
        return 0
    prod: np.ndarray = (series_a - series_a.mean()) * (series_b - series_b.mean())
    return prod.sum() / (std_a * std_b * len(series_a))


def zscore(train_y: np.ndarray, test_y: np.ndarray) -> np.ndarray:
    """
    Estimate to what extend each value in test_y violates
    the normal distribution defined by train_y
    """
    scaler = StandardScaler().fit(train_y.reshape(-1, 1))
    return scaler.transform(test_y.reshape(-1, 1))[:, 0]


def zscore_conf(score: float) -> float:
    """
    Convert z-score into confidence about the hypothesis the score is abnormal
    """
    return 1 - 2 * norm.cdf(-abs(score))


class DecomposableScorer(Scorer):
    """
    Score each node separately
    """

    def score_node(
        self,
        graph: Graph,
        series: Dict[Node, Sequence[float]],
        node: Node,
        data: CaseData,
    ) -> Score:
        """
        Estimate how suspicious a node is
        """
        raise NotImplementedError

    def _score(
        self,
        candidates: Sequence[Node],
        series: Dict[Node, Sequence[float]],
        graph: Graph,
        data: CaseData,
    ):
        results: Dict[Node, Score] = {}
        for node in candidates:
            score = self.score_node(graph, series, node, data)
            if score is not None:
                results[node] = score
        return results

    def score(
        self,
        graph: Graph,
        data: CaseData,
        current: float,
        scores: Dict[Node, Score] = None,
    ) -> Dict[Node, Score]:
        series = data.load_data(graph, current)
        candidates = list(series.keys()) if scores is None else list(scores.keys())

        if self._max_workers >= 2:
            results: Dict[Node, Score] = {}
            with ProcessPoolExecutor(max_workers=self._max_workers) as executor:
                tasks = [
                    executor.submit(
                        require_logging(self._score),
                        candidates=candidates[i :: self._max_workers],
                        series=series,
                        graph=graph,
                        data=data,
                    )
                    for i in range(self._max_workers)
                ]
                for task in as_completed(tasks):
                    results.update(task.result())
        else:
            results = self._score(
                candidates=candidates, series=series, graph=graph, data=data
            )

        if scores is None:
            return results
        return {node: scores[node].update(score) for node, score in results.items()}


class NSigmaScorer(DecomposableScorer):
    """
    Score nodes by n-sigma
    """

    def score_node(
        self,
        graph: Graph,
        series: Dict[Node, Sequence[float]],
        node: Node,
        data: CaseData,
    ) -> Score:
        series_y = np.array(series[node])
        train_y: np.ndarray = series_y[: data.train_window]
        test_y: np.ndarray = series_y[-data.test_window :]
        z_scores = zscore(train_y, test_y)
        z_score = self._aggregator(abs(z_scores))
        score = Score(z_score)
        score["z-score"] = z_score
        return score


class Model:
    """
    A combination of the algorithms
    """

    def __init__(
        self,
        graph_factory: GraphFactory,
        scorers: Sequence[Scorer],
        names: Tuple[str, ...] = None,
    ):
        if not scorers:
            raise ValueError("Please provide at least one scorer")
        self._graph_factory = graph_factory
        self._scorers = scorers
        num_scorers = len(scorers)
        if names is None:
            names = []
        names = list(names) + [None] * (1 + num_scorers - len(names))
        self._name_graph = names[0] or graph_factory.__class__.__name__
        self._names_scorer = [
            obj.__class__.__name__ if name is None else name
            for name, obj in zip(names[1:][:num_scorers], scorers)
        ]
        self._name = "-".join([self._name_graph] + self._names_scorer)

    @staticmethod
    def dump(scores: Dict[Node, Score], filename: str):
        """
        Dump scores into the given file
        """
        data = [
            dict(node=node.asdict(), score=score.asdict())
            for node, score in scores.items()
        ]
        dump_json(filename=filename, data=data)

    @staticmethod
    def load(filename: str) -> Dict[Node, Score]:
        """
        Load scores from the given file
        """
        data: List[Dict[str, dict]] = load_json(filename)
        return {Node(**item["node"]): Score(**item["score"]) for item in data}

    @property
    def name(self) -> str:
        """
        Model name
        """
        return self._name

    def analyze(
        self, data: CaseData, current: float, output_dir: str = None
    ) -> List[Tuple[Node, Score]]:
        """
        Conduct root cause analysis
        """
        # 1. Create a graph
        if output_dir is not None:
            graph_filename = os.path.join(output_dir, f"{self._name_graph}.json")
            graph = None
            if os.path.isfile(graph_filename):
                graph = self._graph_factory.load(graph_filename)
            if graph is None:
                graph = self._graph_factory.create(data=data, current=current)
        else:
            graph = self._graph_factory.create(data=data, current=current)

        # 2. Score nodes
        scores: Dict[Node, Score] = None
        for scorer in self._scorers:
            scores = scorer.score(
                graph=graph, data=data, current=current, scores=scores
            )
        return sorted(scores.items(), key=lambda item: item[1].key, reverse=True)


class Evaluation:
    """
    Evalution results
    """

    def __init__(self, recommendation_num: int = 5):
        self._recommendation_num = recommendation_num
        self._accuracy = {k: 0.0 for k in range(1, recommendation_num + 1)}
        self._ranks: List[List[Node]] = []

    def __call__(self, ranks: Sequence[Node], answers: Set[Node]):
        self._ranks.append(ranks[: self._recommendation_num])
        answer_num = len(answers)
        for k in range(1, self._recommendation_num + 1):
            self._accuracy[k] += len(answers.intersection(ranks[:k])) / answer_num

    @property
    def num(self) -> int:
        """
        Number of cases
        """
        return len(self._ranks)

    def dump(self, filename: str) -> None:
        """
        Dump ranks into the given file
        """
        dump_json(
            filename=filename,
            data=[[node.asdict() for node in ranks] for ranks in self._ranks],
        )

    def load(self, filename: str, answers: Sequence[Set[Node]]) -> None:
        """
        Load ranks from the given file
        """
        self._ranks = []
        report: List[List[dict]] = load_json(filename)
        for ranks, answer in zip(report, answers):
            self(
                [Node(entity=node["entity"], metric=node["metric"]) for node in ranks],
                answer,
            )

    def accuracy(self, k: int) -> float:
        """
        AC@k is the average of accuracy@k among cases

        For each case, accuracy@k = |ranks[:k] \\cap answers| / |answers|
        """
        if k not in self._accuracy or not self._ranks:
            return None
        return self._accuracy[k] / len(self._ranks)

    def average(self, k: int) -> float:
        """
        Avg@k = \\sum_{j=1}^{k} AC@j / k
        """
        if k not in self._accuracy or not self._ranks:
            return None
        return sum(self.accuracy(i) for i in range(1, k + 1)) / k


def _analyze(
    consumer: Queue, model: Model, case: Case, current: float, output_dir: str
):
    try:
        ranks = model.analyze(
            data=case.data,
            current=current,
            output_dir=output_dir,
        )
    except LoadingInvalidGraphException:
        ranks = []
    consumer.put(ranks)


def evaluate(
    model: Model,
    cases: Sequence[Case],
    delay: int = 300,
    output_dir: str = None,
    timeout: int = 3600,
) -> Evaluation:
    """
    Evaluate the composition of Scorers

    delay: the expected interval in seconds to conduct root cause analysis
        after the case is detected
    """
    logger = logging.getLogger(
        ".".join([evaluate.__module__, evaluate.__name__, model.name])
    )
    report = Evaluation()
    if output_dir is not None:
        output_filename = os.path.join(output_dir, f"{model.name}.json")
        if os.path.exists(output_filename):
            report.load(output_filename, [case.answer for case in cases])
            return report
        os.makedirs(output_dir, exist_ok=True)

    for index, case in enumerate(cases):
        logger.debug("Analyze case %d", index)
        case_output_dir = None
        if output_dir is not None:
            case_output_dir = os.path.join(output_dir, str(index))

        consumer = Queue()
        task = Process(
            target=require_logging(_analyze),
            kwargs=dict(
                consumer=consumer,
                model=model,
                case=case,
                current=case.data.detect_time + delay,
                output_dir=case_output_dir,
            ),
        )
        task.start()
        task.join(timeout=timeout)
        if task.is_alive():
            task.terminate()
            logger.warning("Timeout for case %d", index)
            ranks = []
        else:
            try:
                ranks = consumer.get(timeout=10)
            except queue.Empty:
                logger.warning("Unexpected termination for case %d", index)
                ranks = []
        report(ranks=[node for node, _ in ranks], answers=case.answer)
    if output_dir is not None:
        report.dump(output_filename)
    return report



"""
Utilities
"""
from abc import ABC
import dataclasses
from textwrap import indent
from typing import Dict
from typing import Iterable
from typing import Tuple
from typing import Type
from typing import Union

from ...utils import load_json


_ALPHAS = (0.01, 0.05, 0.1, 0.5)
_MAX_CONDS_DIMS = (2, 3, 5, 10, None)
_TAU_MAXS = (0, 1, 2, 3)

_ZERO_TO_ONE = (0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0)
_TRUE_AND_FALSE = (True, False)


@dataclasses.dataclass
class Params(ABC):
    """Parameter template"""

    HELP_MESSAGE = "help"
    IS_ITERABLE = "iterable"
    ABBREVIATION = "abbreviation"
    TARGET = "target"

    def __post_init__(self):
        if self.__class__ is Params:
            raise self._prohibit_instantiate()
        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            if (
                isinstance(field.type, type)
                and issubclass(field.type, Params)
                and isinstance(value, dict)
            ):
                value = field.type(**value)
            elif field.metadata.get(self.IS_ITERABLE, True) and (
                isinstance(value, str) or not isinstance(value, Iterable)
            ):
                value = [value]
            setattr(self, field.name, value)

    @classmethod
    def _prohibit_instantiate(cls):
        return TypeError(f"Can't instantiate abstract class {cls.__name__}")

    @classmethod
    def field_doc(cls, field: dataclasses.Field) -> str:
        """
        Generate doc string for Field

        >>> age: int = dataclasses.field(default=1, metadata={"help": "example"})
        >>> age.name = "age"
        >>> Params.field_doc(age)
        'age: example. [Default: 1]'
        """
        if isinstance(field.type, type) and issubclass(field.type, cls):
            return "\n".join(
                [
                    f"{field.name}: {field.type.__doc__}",
                    indent(cls.dataclass_doc(field.type), " " * 2),
                ]
            )
        help_message = field.metadata.get(cls.HELP_MESSAGE)
        if help_message:
            return f"{field.name}: {help_message}. [Default: {field.default}]"
        return f"{field.name}: [Default: {field.default}]"

    @classmethod
    def dataclass_doc(cls, obj: Type[object]) -> str:
        """
        Generate doc string for a class wrapped by dataclass

        >>> @dataclasses.dataclass
        ... class A:
        ...     method: tuple = dataclasses.field(default=())
        >>> Params.dataclass_doc(A)
        '- method: [Default: ()]'
        """
        return "\n".join(
            [f"- {cls.field_doc(field)}" for field in dataclasses.fields(obj)]
        )


@dataclasses.dataclass
class OptionalParams(Params):
    """
    Optional parameters

    Enable all options by default.
    With any arguments provided, options that are not specified will be disabled.
    """

    def __init__(self, **kwargs):
        # pylint: disable=super-init-not-called
        if kwargs:
            for field in dataclasses.fields(self):
                if field.name in kwargs:
                    setattr(self, field.name, field.type(**kwargs[field.name]))
                else:
                    setattr(self, field.name, None)
        else:
            for field in dataclasses.fields(self):
                setattr(self, field.name, field.type())


@dataclasses.dataclass
class GraphFactoryParams(Params):
    """Parameters for graph factories"""


@dataclasses.dataclass
class PCParams(GraphFactoryParams):
    """Parameters for PC algorithm"""

    alpha: Tuple[float, ...] = dataclasses.field(
        default=_ALPHAS,
        metadata={
            Params.HELP_MESSAGE: "Thresholds for p-value",
            Params.ABBREVIATION: "a",
        },
    )
    max_conds_dim: Tuple[float, ...] = dataclasses.field(
        default=_MAX_CONDS_DIMS,
        metadata={
            Params.HELP_MESSAGE: "The maximum size of condition set",
            Params.ABBREVIATION: "m",
        },
    )
    num_cores: int = dataclasses.field(
        default=1,
        metadata={
            Params.IS_ITERABLE: False,
            Params.HELP_MESSAGE: "The number of cores to be used for "
            "parallel estimation of skeleton num_cores shall be an integer",
        },
    )


@dataclasses.dataclass
class PCGaussParams(PCParams):
    """Parameters for PC algorithm with gauss CI test"""


@dataclasses.dataclass
class PCGSquareParams(PCParams):
    """Parameters for PC algorithm with G-square CI test"""


@dataclasses.dataclass
class PCTSParams(GraphFactoryParams):
    """Parameters for PCTS"""

    alpha: Tuple[float, ...] = dataclasses.field(
        default=_ALPHAS,
        metadata={
            Params.HELP_MESSAGE: "Thresholds for p-value",
            Params.ABBREVIATION: "a",
        },
    )
    max_conds_dim: Tuple[float, ...] = dataclasses.field(
        default=_MAX_CONDS_DIMS,
        metadata={
            Params.HELP_MESSAGE: "The maximum size of condition set",
            Params.ABBREVIATION: "m",
        },
    )
    tau_max: Tuple[int, ...] = dataclasses.field(
        default=_TAU_MAXS,
        metadata={
            Params.HELP_MESSAGE: "The maximum lag considered",
            Params.ABBREVIATION: "t",
        },
    )


@dataclasses.dataclass
class StructuralGraphParams(GraphFactoryParams):
    """Parameters for structural graph construction"""


@dataclasses.dataclass(init=False)
class GraphParams(OptionalParams):
    """Parameters for graph factories"""

    pc_gauss: PCGaussParams = dataclasses.field(
        default_factory=PCGaussParams, metadata={Params.IS_ITERABLE: False}
    )
    pc_gsq: PCGSquareParams = dataclasses.field(
        default_factory=PCGSquareParams, metadata={Params.IS_ITERABLE: False}
    )
    pcts: PCTSParams = dataclasses.field(
        default_factory=PCTSParams, metadata={Params.IS_ITERABLE: False}
    )
    structural: StructuralGraphParams = dataclasses.field(
        default_factory=StructuralGraphParams, metadata={Params.IS_ITERABLE: False}
    )


@dataclasses.dataclass
class ScorerParams(Params):
    """Parameters for scorer with an extra graph fields"""

    graph: GraphParams = dataclasses.field(
        default_factory=GraphParams, metadata={Params.IS_ITERABLE: False}
    )


@dataclasses.dataclass
class NSigmaParams(Params):
    """Parameters for n-sigma"""


@dataclasses.dataclass
class SPOTParams(Params):
    """Parameters for SPOT"""

    risk: Tuple[float, ...] = dataclasses.field(
        default=(1e-2, 1e-3, 1e-4),
        metadata={
            Params.HELP_MESSAGE: "The probability of risk",
            Params.ABBREVIATION: "p",
            Params.TARGET: "proba",
        },
    )


@dataclasses.dataclass(init=False)
class ADParams(OptionalParams):
    """Parameters for anomaly detection scorers"""

    nsigma: NSigmaParams = dataclasses.field(
        default_factory=NSigmaParams, metadata={Params.IS_ITERABLE: False}
    )
    spot: SPOTParams = dataclasses.field(
        default_factory=SPOTParams, metadata={Params.IS_ITERABLE: False}
    )


@dataclasses.dataclass
class DFSParams(ScorerParams):
    """Parameters for DFS-based scorers"""

    detector: ADParams = dataclasses.field(
        default_factory=ADParams, metadata={Params.IS_ITERABLE: False}
    )


@dataclasses.dataclass
class MicroscopeParams(DFSParams):
    """Parameters for Microscope"""


@dataclasses.dataclass
class MicroHECLParams(DFSParams):
    """Parameters for MicroHECL"""

    stop_threshold: Tuple[float, ...] = dataclasses.field(
        default=_ZERO_TO_ONE,
        metadata={
            Params.HELP_MESSAGE: "Threshold for stopping",
            Params.ABBREVIATION: "s",
        },
    )


@dataclasses.dataclass
class RandomWalkParams(ScorerParams):
    """Parameters for random walk-based scorers"""

    rho: Tuple[float, ...] = dataclasses.field(
        default=_ZERO_TO_ONE,
        metadata={
            Params.HELP_MESSAGE: "Back-ward probability",
            Params.ABBREVIATION: "r",
        },
    )


@dataclasses.dataclass
class MicroCauseParams(RandomWalkParams):
    """Parameters for MicroCause"""


@dataclasses.dataclass
class CloudRangerParams(RandomWalkParams):
    """Parameters for CloudRanger"""

    beta: Tuple[float, ...] = dataclasses.field(
        default=_ZERO_TO_ONE,
        metadata={
            Params.HELP_MESSAGE: "For second order random walk",
            Params.ABBREVIATION: "b",
        },
    )


@dataclasses.dataclass
class InvariantNetworkParams(Params):
    """Parameters for invariant network-based scorers"""

    discrete: Tuple[bool, ...] = dataclasses.field(
        default=_TRUE_AND_FALSE,
        metadata={
            Params.HELP_MESSAGE: "Whether to use binary or continous value for correlation",
            Params.ABBREVIATION: "d",
        },
    )
    gamma: Tuple[float, ...] = dataclasses.field(
        default=(0.2, 0.5, 0.8),
        metadata={
            Params.HELP_MESSAGE: "A parameter to balance "
            "the award for neighboring nodes to have similar status scores, and"
            "the penalty of large bias from the initial seeds",
            Params.ABBREVIATION: "c",
        },
    )
    tau: Tuple[float, ...] = dataclasses.field(
        default=(0.1, 1),
        metadata={
            Params.HELP_MESSAGE: "A larger tau typically results in more zeros in e",
            Params.ABBREVIATION: "t",
        },
    )


@dataclasses.dataclass
class ENMFParams(InvariantNetworkParams):
    """Parameters for ENMF"""

    use_softmax: Tuple[bool, ...] = dataclasses.field(
        default=_TRUE_AND_FALSE,
        metadata={
            Params.HELP_MESSAGE: "Whether to use confidence or anomaly score",
            Params.ABBREVIATION: "soft",
        },
    )


@dataclasses.dataclass
class CRDParams(InvariantNetworkParams):
    """Parameters for CRD"""

    num_cluster: Tuple[int, ...] = dataclasses.field(
        default=(2, 3, 4, 5),
        metadata={
            Params.HELP_MESSAGE: "The number of clusters in an invariant network",
            Params.ABBREVIATION: "nc",
        },
    )
    alpha: Tuple[float, ...] = dataclasses.field(
        default=(1.2, 1.5, 2),
        metadata={
            Params.HELP_MESSAGE: "A parameter in the Dirichlet distribution with alpha >= 1",
            Params.ABBREVIATION: "a",
        },
    )
    beta: Tuple[float, ...] = dataclasses.field(
        default=(0.1, 1, 10),
        metadata={
            Params.HELP_MESSAGE: "Balance the object functions for network clustering"
            "and broken score",
            Params.ABBREVIATION: "b",
        },
    )
    learning_rate: Tuple[float, ...] = dataclasses.field(
        default=(0.1, 1), metadata={Params.ABBREVIATION: "lr"}
    )


@dataclasses.dataclass
class CIParams(ScorerParams):
    """Parameters for causal inference-based scorers"""

    tau_max: Tuple[int, ...] = dataclasses.field(
        default=_TAU_MAXS,
        metadata={
            Params.HELP_MESSAGE: "The maximum lag to be considered",
            Params.ABBREVIATION: "t",
        },
    )
    regressor: Tuple[str, ...] = dataclasses.field(
        default=("linear", "svr", "rf", "mdn"),
        metadata={
            Params.HELP_MESSAGE: "Regressor",
        },
    )


@dataclasses.dataclass
class RHTParams(CIParams):
    """Parameters for RHT"""


@dataclasses.dataclass
class RHTDAParams(CIParams):
    """Parameters for RHT-DA"""


@dataclasses.dataclass
class ModelParams(ADParams):
    # pylint: disable=too-many-instance-attributes
    """
    Specify model parameters
    """

    dfs: DFSParams = dataclasses.field(
        default=None, metadata={Params.IS_ITERABLE: False}
    )
    micro_scope: MicroscopeParams = dataclasses.field(
        default=None, metadata={Params.IS_ITERABLE: False}
    )
    micro_hecl: MicroHECLParams = dataclasses.field(
        default=None, metadata={Params.IS_ITERABLE: False}
    )
    micro_cause: MicroCauseParams = dataclasses.field(
        default=None, metadata={Params.IS_ITERABLE: False}
    )
    cloud_ranger: CloudRangerParams = dataclasses.field(
        default=None, metadata={Params.IS_ITERABLE: False}
    )
    enmf: ENMFParams = dataclasses.field(
        default=None, metadata={Params.IS_ITERABLE: False}
    )
    crd: CRDParams = dataclasses.field(
        default=None, metadata={Params.IS_ITERABLE: False}
    )
    rht: RHTParams = dataclasses.field(
        default=None, metadata={Params.IS_ITERABLE: False}
    )
    rht_da: RHTDAParams = dataclasses.field(
        default=None, metadata={Params.IS_ITERABLE: False}
    )

    def __init__(self, params: Union[Dict[str, dict], str] = None):
        # pylint: disable=super-init-not-called
        """
        params shall be dict. The name of a json file is also accepted.
        """
        if params is None:
            params = {}
        elif isinstance(params, str):
            params = load_json(params)
        if not isinstance(params, dict):
            raise ValueError(
                f"ModelParams requires dict or json dict, not {type(params)}"
            )
        super().__init__(**params)


ModelParams.__init__.__doc__ += indent(
    "\n".join(["\nParameters:\n", Params.dataclass_doc(ModelParams)]), " " * 8
)



"""
Regress one node on its parents with an Additive Noise Model
"""
import numpy as np
from sklearn.linear_model._base import LinearModel
from sklearn.linear_model import LinearRegression

from .base import Regressor


class ANMRegressor(Regressor):
    """
    Regress one node on its parents with an Additive Noise Model
    assuming x = f(pa(X)) + e and e follows a normal distribution
    """

    def __init__(self, regressor: LinearModel = None, **kwargs):
        super().__init__(**kwargs)
        self._regressor = regressor if regressor else LinearRegression()

    def _score(
        self,
        train_x: np.ndarray,
        test_x: np.ndarray,
        train_y: np.ndarray,
        test_y: np.ndarray,
    ) -> np.ndarray:
        self._regressor.fit(train_x, train_y)
        train_err: np.ndarray = train_y - self._regressor.predict(train_x)
        test_err: np.ndarray = test_y - self._regressor.predict(test_x)
        return self._zscore(train_y=train_err, test_y=test_err)



"""
Command line utilities
"""
import argparse
import dataclasses
from enum import Enum
import json
import logging
import os
from typing import Any
from typing import List
from typing import Tuple

import numpy as np

from . import comparison
from .comparison.utils import ModelParams
from .simulation import SimDataset
from .simulation import generate
from .simulation import robustness
from .simulation.scorer import SimRHTScorer
from ..alg.common import Model
from ..utils import silence_third_party


_GRAPH_SIZES: List[Tuple[int, int]] = [
    (50, 100),
    (100, 500),
    (500, 5000),
]
_NUM_NODES = [n for n, _ in _GRAPH_SIZES]
_NUM_GRAPHS = 10
_DEFAULT_SEED = 519


def _show_params(_: argparse.Namespace):
    def _convert_value(obj):
        if isinstance(obj, Enum):
            return obj.value
        if isinstance(obj, (list, tuple)):
            return list(map(_convert_value, obj))
        return obj

    def _enum2str(data: List[Tuple[str, Any]]) -> dict:
        return {key: _convert_value(value) for key, value in data}

    params = dataclasses.asdict(ModelParams(), dict_factory=_enum2str)
    print(json.dumps(params, indent=2, ensure_ascii=False))


def _generate(args: argparse.Namespace):
    num_nodes = set(args.num_nodes)
    rng = np.random.default_rng(args.seed)
    for num_node, num_edge in _GRAPH_SIZES:
        if num_node not in num_nodes:
            continue
        for i in range(_NUM_GRAPHS):
            dataset = generate(
                num_node=num_node, num_edge=num_edge, num_cases=args.num_cases, rng=rng
            )
            dataset.dump(os.path.join(args.output_dir, str(num_node), str(i)))


def _tune(args: argparse.Namespace):
    from .comparison.models import get_models  # pylint: disable=import-outside-toplevel

    data_dir: str = args.data_dir
    report_dir: str = args.report_dir
    os.makedirs(report_dir, exist_ok=True)

    logger = logging.getLogger(__package__)
    case_data_params = dict(lookup_window=120)  # 2 hours

    logger.info("Loading from %s", data_dir)
    dataset = SimDataset.load(
        folder=data_dir,
        case_data_params=case_data_params,
    )
    models, _ = get_models(
        graph_factories={"GT": dataset.graph_factory},  # GT: Ground Truth
        params=args.model_params,
        seed=args.seed,
        cuda=args.cuda,
    )
    logger.info("Start tuning on %s with #models=%d", data_dir, len(models))
    comparison.run(
        models=models,
        cases=dataset.cases,
        graph_factories=None,
        output_dir=args.output_dir,
        report_filename=os.path.join(report_dir, "report.csv"),
        max_workers=1 if args.cuda else args.max_workers,
    )


def _run(args: argparse.Namespace):
    from .comparison.models import get_models  # pylint: disable=import-outside-toplevel

    model_params: ModelParams = args.model_params
    report_dir: str = args.report_dir
    os.makedirs(report_dir, exist_ok=True)

    logger = logging.getLogger(__package__)
    case_data_params = dict(lookup_window=120)

    for num_node, _ in _GRAPH_SIZES:
        for i in range(_NUM_GRAPHS):
            dataset_dir = os.path.join(args.data_dir, str(num_node), str(i))
            logger.info("Loading from %s", dataset_dir)
            dataset = SimDataset.load(
                folder=dataset_dir,
                case_data_params=case_data_params,
            )
            models, _ = get_models(
                graph_factories={"GT": dataset.graph_factory},  # GT: Ground Truth
                params=model_params,
                seed=args.seed,
                cuda=args.cuda,
                max_workers=args.max_workers,
            )
            if model_params.rht:
                models.append(
                    Model(
                        graph_factory=dataset.graph_factory,
                        scorers=[
                            SimRHTScorer(seed=args.seed, max_workers=args.max_workers)
                        ],
                        names=["GT", "RHT-PG"],
                    ),
                )
            logger.info("Start running on %s", dataset_dir)
            comparison.run(
                models=models,
                cases=dataset.cases,
                graph_factories=None,
                output_dir=os.path.join(args.output_dir, str(num_node), str(i)),
                report_filename=os.path.join(report_dir, f"report-{num_node}-{i}.csv"),
                max_workers=1,
            )


def _run_robustness(args: argparse.Namespace):
    cache_dir: str = args.cache_dir
    data_dir: str = args.data_dir
    report_dir: str = args.report_dir
    os.makedirs(report_dir, exist_ok=True)

    for num_node, _ in _GRAPH_SIZES:
        robustness.evaluate(
            num_graph=_NUM_GRAPHS,
            cache_dir=os.path.join(cache_dir, str(num_node)),
            data_dir=os.path.join(data_dir, str(num_node)),
            report_dir=os.path.join(report_dir, str(num_node)),
        )


def _add_output_argument(parser: argparse.ArgumentParser, default: str):
    parser.add_argument(
        "--output-dir", type=str, default=default, help="Output directory"
    )


def _add_report_argument(parser: argparse.ArgumentParser, default: str):
    parser.add_argument(
        "--report-dir", type=str, default=default, help="Report directory"
    )


def get_parser() -> Tuple[argparse.ArgumentParser, argparse._SubParsersAction]:
    """
    Prepare the command line parser
    """
    parser_params = dict(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    # The entrance
    parser = argparse.ArgumentParser(prog=__package__, **parser_params)
    parser.add_argument("-v", dest="V", action="store_true", help="Verbose")
    parser.add_argument("-s", dest="S", action="store_true", help="Silence")
    parser.add_argument("--seed", type=int, default=_DEFAULT_SEED, help="Random seed")
    parser.add_argument(
        "--max-workers",
        type=int,
        default=1,
        help="The number of workers for parallel calculation",
    )
    parser.add_argument("--cuda", action="store_true", help="Use CUDA (GPU)")
    parser.add_argument(
        "--model-params",
        type=ModelParams,
        required=False,
        help="Provide a json file to specify options for model parameters.",
    )
    subparsers = parser.add_subparsers(title="subcommands")

    parser_show_params = subparsers.add_parser(
        "params", help="Show default model parameters", **parser_params
    )
    parser_show_params.set_defaults(func=_show_params)

    # For data generation
    parser_gen = subparsers.add_parser(
        "generate", help="Generate dataset by simulation", **parser_params
    )
    parser_gen.add_argument(
        "--num-cases",
        type=int,
        default=100,
        help="The number of cases for each graph",
    )
    parser_gen.add_argument(
        "-n",
        "--num-node",
        dest="num_nodes",
        type=int,
        choices=_NUM_NODES,
        default=_NUM_NODES,
        nargs="*",
        help="Choose the number of nodes in the generated graph",
    )
    _add_output_argument(parser_gen, default="dataset")
    parser_gen.set_defaults(func=_generate)

    # Experiments over parameter combinations
    parser_tune = subparsers.add_parser(
        "tune",
        help="Explore all combinations of model parameters in the given dataset",
        **parser_params,
    )
    parser_tune.add_argument(
        "--data-dir",
        type=str,
        default=os.path.join("dataset", str(_GRAPH_SIZES[0][0]), "0"),
        help="Data directory",
    )
    _add_output_argument(parser_tune, default=os.path.join("output", "sim-tuning"))
    _add_report_argument(parser_tune, default=os.path.join("report", "sim-tuning"))
    parser_tune.set_defaults(func=_tune)

    # Experiments on a set of parameters
    parser_run = subparsers.add_parser(
        "run",
        help="Explore all the dataset with pre-defined model parameters",
        **parser_params,
    )
    parser_run.add_argument(
        "--data-dir",
        type=str,
        default="dataset",
        help="Data directory",
    )
    _add_output_argument(parser_run, default=os.path.join("output", "sim"))
    _add_report_argument(parser_run, default=os.path.join("report", "sim"))
    parser_run.set_defaults(func=_run)

    parser_robustness = subparsers.add_parser(
        "robustness",
        help="Compare models with faults of different dependency intensities",
        **parser_params,
    )
    parser_robustness.add_argument(
        "--data-dir",
        type=str,
        default=os.path.join("dataset"),
        help="Data directory",
    )
    parser_robustness.add_argument(
        "--cache-dir",
        type=str,
        default=os.path.join("output", "sim"),
        help="Data directory",
    )
    _add_report_argument(
        parser_robustness, default=os.path.join("report", "sim-robustness")
    )
    parser_robustness.set_defaults(func=_run_robustness)

    return parser, subparsers


def _main():
    parser, _ = get_parser()
    parameters = parser.parse_args()

    if parameters.S:
        logging.basicConfig(level=logging.ERROR)
    elif parameters.V:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)
    if not parameters.V:
        silence_third_party()

    if "func" in parameters:
        parameters.func(parameters)
    else:
        parser.print_usage()


if __name__ == "__main__":
    _main()



"""
Ranking Causal Anomalies via Temporal and Dynamical Analysis on Vanishing Correlations

See KDD'16
https://doi.org/10.1145/2939672.2939765
"""
from itertools import combinations
import logging
from typing import Dict
from typing import Tuple

import numpy as np
from scipy.special import softmax


_ZERO = 1e-6


class ARX:
    # pylint: disable=invalid-name, too-many-instance-attributes
    """
    AutoRegressive eXogenous (ARX) model

    y(t) = \\sum_{i=1}^{n} a_{i} y(t - i) + \\sum_{i=0}^{m} b_{i} x(t - k - i) + d

    where 0 <= n, m, l <= k is a popular choice
    """

    def __init__(
        self,
        train_x: np.ndarray,
        train_y: np.ndarray,
        moment_2: float = None,
        n: int = 3,
        m: int = 2,
        k: int = 0,
    ):
        # pylint: disable=too-many-arguments
        """
        Fit the ARX model with the training data
        """
        self._n = n
        self._m = m
        self._k = k
        self._start = max(self._n, self._m + self._k)

        x, y = self._prepare_data(train_x=train_x, train_y=train_y)
        theta: np.ndarray = np.linalg.lstsq(x, y, rcond=None)[0]
        residuals: np.ndarray = y - x @ theta

        self._theta = theta
        self._epsilon_max: float = abs(residuals).max()

        if moment_2 is None:
            moment_2 = y.shape[0] * y.var()
        self._fitness: float = 1 - np.sqrt((residuals**2).sum() / moment_2)

    @property
    def epsilon_max(self) -> float:
        """
        Maximum training error
        """
        return self._epsilon_max

    @property
    def fitness(self) -> float:
        """
        Fitness on the training data
        """
        return self._fitness

    def _prepare_data(self, train_x: np.ndarray, train_y: np.ndarray):
        num_samples = train_y.size
        y: np.ndarray = train_y[self._start :]
        x = np.array(
            [train_y[self._start - i : -i] for i in range(1, self._n + 1)]
            + [
                train_x[self._start - self._k - i : num_samples - self._k - i]
                for i in range(self._m + 1)
            ]
            + [np.ones(train_y.size - self._start)]
        ).T
        return x, y

    def residuals(self, test_x: np.ndarray, test_y: np.ndarray) -> np.ndarray:
        """
        Residuals on the testing data
        """
        x, y = self._prepare_data(train_x=test_x, train_y=test_y)
        return y - x @ self._theta


class InvariantNetwork:
    """
    Learn invariant network based on the AutoRegressive eXogenous (ARX) model
    """

    LOWER_THRESHOLD = 0.5
    UPPER_THRESHOLD = 0.7

    BROKEN_THRESHOLD = 1.1

    def __init__(self, n: int = 3, m: int = 2, k: int = 0):
        # pylint: disable=invalid-name
        """
        (n, m) is the order of the ARX model
        """
        self._arx_params = dict(n=n, m=m, k=k)
        self._start = max(n, m + k)
        self._edges: Dict[Tuple[int, int], ARX] = None
        self._num_nodes = 0

    @property
    def start(self) -> int:
        """
        The number of samples that will not be predicted by the ARX model,
            which is max(n, m + k)
        """
        return self._start

    def fit(self, data: np.ndarray):
        """
        Search invariant in a brute-force way

        The judgement of invariant edge comes from the code[2] of Reference[1] in KDD'16

        References:
        [1] Ranking Causal Anomalies via Temporal and Dynamical Analysis on Vanishing
            Correlations. https://doi.org/10.1145/2939672.2939765
        [2] gRank-mRank.zip:gRank-mRank/code/pruning/Scratch.m in
            https://github.com/chengw07/CausalRanking
        """
        edges = {}
        self._num_nodes = data.shape[1]

        valid_data: np.ndarray = data[self._start :, :]
        moment_2: np.ndarray = valid_data.shape[0] * valid_data.var(axis=0)
        for i, j in combinations(np.where(moment_2)[0], 2):
            arx_ij = ARX(
                train_x=data[:, j],
                train_y=data[:, i],
                moment_2=moment_2[i],
                **self._arx_params,
            )
            arx_ji = ARX(
                train_x=data[:, i],
                train_y=data[:, j],
                moment_2=moment_2[j],
                **self._arx_params,
            )
            if (
                min(arx_ij.fitness, arx_ji.fitness) >= self.LOWER_THRESHOLD
                and max(arx_ij.fitness, arx_ji.fitness) >= self.UPPER_THRESHOLD
            ):
                if arx_ji.fitness >= arx_ij.fitness:
                    edges[(j, i)] = arx_ji
                else:
                    edges[(i, j)] = arx_ij

        self._edges = edges

    def invariant_matrix(self, discrete: bool = True) -> np.ndarray:
        """
        The symmetric matrix of the invariant network
        """
        matrix = np.zeros((self._num_nodes, self._num_nodes), dtype=float)
        if discrete:
            for i, j in self._edges:
                matrix[i, j] = matrix[j, i] = 1
        else:
            for (i, j), arx in self._edges.items():
                matrix[i, j] = matrix[j, i] = arx.fitness
        return matrix

    def broken_matrix(self, data: np.ndarray, discrete: bool = True) -> np.ndarray:
        """
        The symmetric matrix of the broken network

        Notice:
            The code[3] of Reference[1] in KDD'16 uses 1.1 times of
            the 99.5 percentiles of training residuals as the threshold

        References:
        [1] Ranking Causal Anomalies via Temporal and Dynamical Analysis on Vanishing
            Correlations. https://doi.org/10.1145/2939672.2939765
        [3] gRank-mRank.zip:gRank-mRank/code/ranking(realdata)/InvTracking.m in
            https://github.com/chengw07/CausalRanking
        """
        assert data.shape[1] == self._num_nodes
        matrix = np.zeros((self._num_nodes, self._num_nodes), dtype=float)
        for (i, j), arx in self._edges.items():
            residuals = arx.residuals(test_x=data[:, j], test_y=data[:, i])
            normalized_residual = np.nanmax(abs(residuals) / arx.epsilon_max)
            if normalized_residual > self.BROKEN_THRESHOLD:
                matrix[i, j] = matrix[j, i] = 1 if discrete else normalized_residual
        return matrix


class ENMF:
    """
    ENMF is edited based on the code[4] of Reference[1] in KDD'16

    References:
    [1] Ranking Causal Anomalies via Temporal and Dynamical Analysis on Vanishing
        Correlations. https://doi.org/10.1145/2939672.2939765
    [4] rankingCausal.zip:rankingCausal/optENMF.m in
        https://github.com/chengw07/CausalRanking
    """

    def __init__(
        self,
        gamma: float = 0.5,
        tau: float = 1.0,
        epoches: int = 500,
        invariant_network: InvariantNetwork = None,
    ):
        """
        Constructor

        Parameters:
            gamma: A parameter to balance
                the award for neighboring nodes to have similar status scores, and
                the penalty of large bias from the initial seeds
            tau: A larger tau typically results in more zeros in e
        """
        self._gamma = gamma
        self._tau = tau

        self._epoches = epoches

        if invariant_network is None:
            self._invariant_network = InvariantNetwork()
        else:
            self._invariant_network = invariant_network

        self._logger = logging.getLogger(
            f"{self.__class__.__module__}.{self.__class__.__name__}"
        )

    @staticmethod
    def degree_normalize(matrix: np.ndarray):
        """
        Calculate D^{-0.5} A D^{-0.5},
        where D is a diagonal matrix with D_{xx} = \\sum_{y} A_{xy}
        """
        d_root = np.diag(
            [0 if item == 0 else 1 / item for item in np.sqrt(matrix.sum(axis=1))]
        )
        return d_root * matrix * d_root

    @property
    def start(self) -> int:
        """
        The number of samples that will not be predicted by the ARX model,
            which is max(n, m + k)
        """
        return self._invariant_network.start

    def _loss(
        self,
        broken: np.ndarray,
        causes: np.ndarray,
        status_scores: np.ndarray,
        mask: np.ndarray,
    ) -> Tuple[float, float]:
        # Square of Frobenius norm
        reconstruct_loss: float = (
            np.linalg.norm(
                mask * (status_scores @ status_scores.T) - broken,
                ord="fro",
            )
            ** 2
        )
        loss: float = reconstruct_loss + self._tau * abs(causes).sum()
        return loss, reconstruct_loss

    def enmf(
        self,
        invariant: np.ndarray,
        broken: np.ndarray,
        causes: np.ndarray,
        weight: np.ndarray = None,
    ):
        # pylint: disable=too-many-locals
        """
        Causal Anomaly Ranking with an iterative updating algorithm
        """
        # A: invariant network
        # B: broken network
        # e: causes
        # H: propagation with r = H e, where r is the anomaly score vector
        # H = (1 - c) (I_n - c \tilde{A})^{-1}
        propagation: np.ndarray = (1 - self._gamma) * np.linalg.inv(
            np.eye(invariant.shape[0]) - self._gamma * self.degree_normalize(invariant)
        )
        mask = invariant.astype(bool).astype(float)

        def _status_scores(causes: np.ndarray) -> np.ndarray:
            if weight is None:
                return propagation @ causes
            return propagation @ (weight * causes)

        for epoch in range(self._epoches):
            causes_old = causes
            status_scores = _status_scores(causes)
            numerator: np.ndarray = (
                4 * (mask * (propagation.T @ broken)) @ status_scores
            )
            denominator: np.ndarray = (
                4
                * (mask * (propagation.T @ status_scores @ status_scores.T))
                @ status_scores
                + self._tau
            )
            causes = causes * ((numerator / denominator) ** 0.25)
            err: float = np.linalg.norm(causes - causes_old, ord="fro")

            loss, reconstruct_loss = self._loss(
                broken=broken,
                causes=causes,
                status_scores=_status_scores(causes),
                mask=mask,
            )

            self._logger.debug(
                "Epoch of ENMF: %04d, loss: %06f, reconstruct_loss: %06f, change: %06f",
                epoch,
                loss,
                reconstruct_loss,
                err,
            )
            if err < _ZERO:
                break
        return causes

    def train(self, data: np.ndarray):
        """
        Learn the invariant network
        """
        self._invariant_network.fit(data)

    def _rank(self, invariant: np.ndarray, broken: np.ndarray) -> np.ndarray:
        causes = np.ones((invariant.shape[0], 1))
        causes = self.enmf(invariant=invariant, broken=broken, causes=causes)
        return causes[:, 0]

    def rank(self, data: np.ndarray, discrete: bool = True) -> np.ndarray:
        """
        Detect the broken network and rank nodes

        Returns:
            An array of causal anomaly scores for each node
        """
        invariant = self._invariant_network.invariant_matrix(discrete=discrete)
        if np.all(invariant == 0):
            return np.zeros(invariant.shape[0])
        broken = self._invariant_network.broken_matrix(data=data, discrete=discrete)

        return self._rank(invariant=invariant, broken=broken)


class ENMFSoft(ENMF):
    """
    ENMF with softmax normalization

    ENMFSoft is edited based on the code[5] of Reference[1] in KDD'16

    References:
    [1] Ranking Causal Anomalies via Temporal and Dynamical Analysis on Vanishing
        Correlations. https://doi.org/10.1145/2939672.2939765
    [5] rankingCausal.zip:rankingCausal/optENMFSoft.m in
        https://github.com/chengw07/CausalRanking
    """

    def _rank(self, invariant: np.ndarray, broken: np.ndarray) -> np.ndarray:
        # pylint: disable=too-many-locals
        # A: invariant network
        # B: broken network
        # e: causes
        # H: propagation with r = H e, where r is the anomaly score vector
        # H = (1 - c) (I_n - c \tilde{A})^{-1}
        propagation: np.ndarray = (1 - self._gamma) * np.linalg.inv(
            np.eye(invariant.shape[0]) - self._gamma * self.degree_normalize(invariant)
        )
        mask = invariant.astype(bool).astype(float)

        causes = np.ones((invariant.shape[0], 1))
        status_scores = softmax(propagation @ causes)
        loss, _ = self._loss(
            broken=broken, causes=causes, status_scores=status_scores, mask=mask
        )

        for epoch in range(self._epoches):
            loss_old = loss
            status_scores: np.ndarray = softmax(propagation @ causes)
            phi = np.diag(status_scores) - status_scores @ status_scores.T

            numerator: np.ndarray = (
                4 * (mask * (propagation.T @ phi @ broken)) @ status_scores
            )
            numerator[numerator < 0] = 0
            denominator: np.ndarray = (
                4
                * (mask * (propagation.T @ phi @ status_scores @ status_scores.T))
                @ status_scores
                + self._tau
            )
            causes = causes * ((numerator / denominator) ** 0.25)
            loss, reconstruct_loss = self._loss(
                broken=broken, causes=causes, status_scores=status_scores, mask=mask
            )
            err = abs(loss - loss_old)

            self._logger.debug(
                "Epoch of ENMF: %04d, loss: %06f, reconstruct_loss: %06f, change: %06f",
                epoch,
                loss,
                reconstruct_loss,
                err,
            )
            if err < _ZERO:
                break
        return causes[:, 0]



"""
Define the interface for algorithms to access relations
"""
from abc import ABC
from typing import Dict
from typing import List
from typing import Set
from typing import Union

import networkx as nx

from ..utils import dump_json
from ..utils import load_json
from ..utils import topological_sort


class Node:
    """
    The element of a graph
    """

    def __init__(self, entity: str, metric: str):
        self._entity = entity
        self._metric = metric

    @property
    def entity(self) -> str:
        """
        Entity getter
        """
        return self._entity

    @property
    def metric(self) -> str:
        """
        Metric getter
        """
        return self._metric

    def asdict(self) -> Dict[str, str]:
        """
        Serialized as a dict
        """
        return {"entity": self._entity, "metric": self._metric}

    def __eq__(self, obj: object) -> bool:
        if isinstance(obj, Node):
            return self.entity == obj.entity and self.metric == obj.metric
        return False

    def __hash__(self) -> int:
        return hash((self.entity, self.metric))

    def __repr__(self) -> str:
        return f"Node{(self.entity, self.metric)}"


class LoadingInvalidGraphException(Exception):
    """
    This exception indicates that Graph tries to load from a broken file
    """


class Graph(ABC):
    """
    The abstract interface to access relations
    """

    def __init__(self):
        self._nodes: Set[Node] = set()
        self._sorted_nodes: List[Set[Node]] = None

    def dump(self, filename: str) -> bool:
        # pylint: disable=no-self-use, unused-argument
        """
        Dump a graph into the given file

        Return whether the operation succeeds
        """
        return False

    @classmethod
    def load(cls, filename: str) -> Union["Graph", None]:
        # pylint: disable=unused-argument
        """
        Load a graph from the given file

        Returns:
        - A graph, if available
        - None, if dump/load is not supported
        - Raise LoadingInvalidGraphException if the file cannot be parsed
        """
        return None

    @property
    def nodes(self) -> Set[Node]:
        """
        Get the set of nodes in the graph
        """
        return self._nodes

    @property
    def topological_sort(self) -> List[Set[Node]]:
        """
        Sort nodes with parents first

        The graph specifies the parents of each node.
        """
        if self._sorted_nodes is None:
            self._sorted_nodes = topological_sort(
                nodes=self.nodes, predecessors=self.parents, successors=self.children
            )
        return self._sorted_nodes

    def children(self, node: Node, **kwargs) -> Set[Node]:
        """
        Get the children of the given node in the graph
        """
        raise NotImplementedError

    def parents(self, node: Node, **kwargs) -> Set[Node]:
        """
        Get the parents of the given node in the graph
        """
        raise NotImplementedError


class MemoryGraph(Graph):
    """
    Implement Graph with data in memory
    """

    def __init__(self, graph: nx.DiGraph):
        """
        graph: The whole graph
        """
        super().__init__()
        self._graph = graph
        self._nodes.update(self._graph.nodes)

    def dump(self, filename: str) -> bool:
        nodes: List[Node] = list(self._graph.nodes)
        node_indexes = {node: index for index, node in enumerate(nodes)}
        edges = [
            (node_indexes[cause], node_indexes[effect])
            for cause, effect in self._graph.edges
        ]
        data = dict(nodes=[node.asdict() for node in nodes], edges=edges)
        dump_json(filename=filename, data=data)

    @classmethod
    def load(cls, filename: str) -> Union["MemoryGraph", None]:
        data: dict = load_json(filename=filename)
        if "nodes" not in data or "edges" not in data:
            raise LoadingInvalidGraphException(filename)
        nodes: List[Node] = [Node(**node) for node in data["nodes"]]
        graph = nx.DiGraph()
        graph.add_nodes_from(nodes)
        graph.add_edges_from(
            (nodes[cause], nodes[effect]) for cause, effect in data["edges"]
        )
        return MemoryGraph(graph)

    def children(self, node: Node, **kwargs) -> Set[Node]:
        if not self._graph.has_node(node):
            return set()
        return set(self._graph.successors(node))

    def parents(self, node: Node, **kwargs) -> Set[Node]:
        if not self._graph.has_node(node):
            return set()
        return set(self._graph.predecessors(node))



"""
Regress one node on its parents with the Gaussian Mixture Model
"""
import numpy as np

from .base import GMMPredictor
from .prob_rf import ProbRF
from ..base import Regressor


class GMMRegressor(Regressor):
    """
    Regress one node on its parents with the Gaussian Mixture Model
    """

    def __init__(
        self,
        regressor: GMMPredictor = None,
        sample_size: int = 100,
        seed: int = None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self._regressor = regressor if regressor else ProbRF(seed=seed)
        self._sample_size = sample_size

    def _score(
        self,
        train_x: np.ndarray,
        test_x: np.ndarray,
        train_y: np.ndarray,
        test_y: np.ndarray,
    ) -> np.ndarray:
        self._regressor.train(train_x, train_y)
        return np.array(
            [
                self._zscore(gmm.sample(self._sample_size), target)[0]
                for gmm, target in zip(self._regressor.predict(test_x), test_y)
            ]
        )



"""
Wrapper for R tools
"""
from enum import Enum
import os
from typing import List
from typing import Tuple

import numpy as np
from rpy2 import robjects
from rpy2.robjects import numpy2ri
from rpy2.robjects import default_converter
from rpy2.robjects.conversion import localconverter

from .utils import bcp
from ..common import DynamicGraphFactory
from ...model.graph import Node


robjects.r(
    f"source(\"{os.path.dirname(__file__).replace(os.path.sep, '/')}/causal.R\")"
)


class PCAlgFactory(DynamicGraphFactory):
    """
    Mine the graph with the R package pcalg
    """

    class _Method(Enum):

        PC_GAUSS = "PC-gauss"
        PC_GSQ = "PC-gsq"

    def __init__(
        self,
        alpha: float = 0.05,
        method: str = _Method.PC_GAUSS.value,
        tau_max: int = 3,
        max_conds_dim: int = np.inf,
        num_cores: int = 1,
        **kwargs,
    ):
        # pylint: disable=too-many-arguments
        """
        alpha: desired significance level in (0, 1)
        method: name of the method to be used
            - "PC-gauss": PC algorithm with gaussCItest
            - "PC-gsq": PC algorithm with disCItest
        max_conds_dim: Maximum size of a condition set, unrestricted by default.
        """
        super().__init__(**kwargs)
        self._alpha = alpha
        assert method in {_method.value for _method in self._Method}
        self._method = self._Method(method)
        self._tau_max = tau_max
        self._max_conds_dim = max_conds_dim
        self._num_cores = num_cores

    def _create(
        self, data: np.ndarray, nodes: List[Node]
    ) -> Tuple[np.ndarray, List[Node]]:
        if self._method == self._Method.PC_GSQ:
            series = {
                node: bcp(data[:, index], tau_max=self._tau_max)
                for index, node in enumerate(nodes)
            }
            nodes = [
                node
                for node, change_points in series.items()
                if len(np.unique(change_points)) > 1
            ]
            if len(nodes) < 2:
                return np.zeros((0, 0)), nodes
            params = {
                "d": np.array([series[node] for node in nodes]).T,
                "CItest": "gsq",
                "alpha": self._alpha,
                "m.max": self._max_conds_dim,
                "numCores": self._num_cores,
            }
            fun = robjects.r["runPC"]
        else:  # self._method == self._Method.PC_GAUSS
            params = {
                "d": data,
                "CItest": "gauss",
                "alpha": self._alpha,
                "m.max": self._max_conds_dim,
                "numCores": self._num_cores,
            }
            fun = robjects.r["runPC"]

        with localconverter(default_converter + numpy2ri.converter):
            matrix: np.ndarray = fun(**params)
        return matrix, nodes



"""
Utilities
"""
from functools import partial

import bayesian_changepoint_detection.online_changepoint_detection as oncd
import numpy as np


def bcp(data: np.ndarray, tau_max: int = 3, confidence: float = 0.5) -> np.ndarray:
    """
    Bayesian change-point detection

    Used in CauseInfer in INFOCOM'14
    https://doi.org/10.1109/INFOCOM.2014.6848128
    """
    probs, _ = oncd.online_changepoint_detection(
        data, partial(oncd.constant_hazard, 250), oncd.StudentT(0.1, 0.01, 1, 0)
    )
    change_points = np.zeros(data.shape)
    change_points[tau_max:] = probs[tau_max, tau_max + 1 :] > confidence
    return change_points



"""
Predict one node on its parents with the Gaussian Mixture Model
"""
from abc import ABC
from typing import List
from typing import Sequence

import numpy as np
from sklearn.mixture import GaussianMixture


class GMM(ABC):
    """
    Gaussian Mixture Model
    """

    def sample(self, n_samples: int = 1) -> np.ndarray:
        """
        Sample from the model and return 1-D array
        """
        raise NotImplementedError


class SKLearnGMM(GMM):
    """
    GMM with sklearn
    """

    def __init__(self, samples: Sequence[float], gmm_k: int = 2, seed: int = None):
        self._model = GaussianMixture(
            n_components=min(len(samples), gmm_k),
            covariance_type="full",
            random_state=seed,
        )
        self._model.fit(samples)

    def sample(self, n_samples: int = 1) -> np.ndarray:
        samples = self._model.sample(n_samples=n_samples)[0]
        return samples.reshape(-1)


class GMMPredictor(ABC):
    """
    Predict one node on its parents with the Gaussian Mixture Model
    """

    def __init__(self, gmm_k: int = 2, seed: int = None):
        self._gmm_k = gmm_k
        self._seed = seed

    def train(self, train_x: np.ndarray, train_y: np.ndarray):
        """
        Train the model
        """
        raise NotImplementedError

    def predict(self, test_x: np.ndarray) -> List[GMM]:
        """
        Generate GMM
        """
        raise NotImplementedError



"""
Define the data structure for algorithms as context
"""
import datetime
from typing import Dict
from typing import Sequence
from typing import Set

import numpy as np

from .data_loader import DataLoader
from .graph import Graph
from .graph import Node


class CaseData:
    # pylint: disable=too-many-instance-attributes
    """
    Case data that algorithms can access
    """

    def __init__(
        self,
        data_loader: DataLoader,
        sli: Node,
        detect_time: float,
        interval: datetime.timedelta = datetime.timedelta(minutes=1),
        lookup_window: int = 120,
        detect_window: int = 10,
        prune: bool = True,
    ):
        # pylint: disable=too-many-arguments
        self._data_loader = data_loader
        self._sli = sli
        self._detect_time = detect_time

        # Parameters for the algorithm
        self._interval = interval
        self._train_window = lookup_window - detect_window + 1
        self._test_window = detect_window
        self._lookup_window = lookup_window * interval.total_seconds()
        self._prune = prune

    @property
    def data_loader(self) -> DataLoader:
        """
        Interface to access raw data
        """
        return self._data_loader

    @property
    def sli(self) -> Node:
        """
        The service level indicator (SLI) that is violated
        """
        return self._sli

    @property
    def detect_time(self) -> float:
        """
        Unix timestamp when the service level indicator (SLI) is violated
        """
        return self._detect_time

    @property
    def train_window(self) -> int:
        """
        Number of data points for learning the normal pattern
        """
        return self._train_window

    @property
    def test_window(self) -> int:
        """
        Number of data points for analyzing the fault
        """
        return self._test_window

    def load_data(
        self, graph: Graph = None, current: float = None
    ) -> Dict[Node, Sequence[float]]:
        """
        Parepare data
        """
        if current is None:
            current = self._detect_time
        else:
            current = max(current, self._detect_time)
        nodes = self._data_loader.nodes if graph is None else graph.nodes

        start = self._detect_time - self._lookup_window
        length = int((current - start) / self._interval.total_seconds()) + 1
        series: Dict[Node, Sequence[float]] = {}
        for node in nodes:
            node_data = self._data_loader.load(
                entity=node.entity,
                metric=node.metric,
                start=start,
                end=current,
                interval=self._interval,
            )
            if self._prune:
                if node_data and len(set(node_data)) > 1:
                    series[node] = node_data[:length]
            else:
                if not node_data:
                    node_data = np.zeros(length)
                series[node] = node_data[:length]
        return series


class Case:
    """
    Case data for evaluation
    """

    def __init__(self, data: CaseData, answer: Set[Node]):
        self._data = data
        self._answer = answer

    @property
    def data(self) -> CaseData:
        """
        Case data
        """
        return self._data

    @property
    def answer(self) -> Set[Node]:
        """
        Ground truth for this case
        """
        return self._answer



"""
Wrapper for PCMCI
"""
import logging
from typing import List
from typing import Tuple

import numpy as np
from tigramite import data_processing
from tigramite.independence_tests import ParCorr
from tigramite.independence_tests.independence_tests_base import CondIndTest
from tigramite.pcmci import PCMCI

from .common import DynamicGraphFactory
from ..model.graph import Node


class _ParCorr(ParCorr):
    # pylint: disable=invalid-name
    """
    Wrap ParCorr to handle constant
    """

    _logger_name = f"{ParCorr.__module__}.{ParCorr.__name__}"

    def _get_single_residuals(
        self,
        array: np.ndarray,
        target_var: int,
        standardize: bool = True,
        return_means: bool = False,
    ) -> np.ndarray:
        y: np.ndarray = array[target_var, :]
        z: np.ndarray = np.copy(array[2:, :])

        # Standardize
        if standardize:
            y -= y.mean()
            std: np.ndarray = y.std()
            if std > 0:
                y /= std

            z -= z.mean(axis=1).reshape(-1, 1)
            std = z.std(axis=1)
            # Skip constant variables
            indexes: np.ndarray = np.where(std)[0]
            z = z[indexes, :] / std[indexes].reshape(-1, 1)
            if np.isnan(array).sum() != 0:
                raise ValueError("nans after standardizing, possibly constant array!")

        if z.shape[0] > 0:
            z = z.T
            try:
                beta_hat = np.linalg.lstsq(z, y, rcond=None)[0]
                mean = np.dot(z, beta_hat)
                resid = y - mean
            except np.linalg.LinAlgError as err:
                logging.getLogger(self._logger_name).warning(err, exc_info=True)
                resid = y
                mean = None
        else:
            resid = y
            mean = None

        if return_means:
            return (resid, mean)
        return resid


class PCTSFactory(DynamicGraphFactory):
    """
    Mine the graph with path condition time series (PCTS [1]) based on PCMCI [2]

    References
    [1] MicroCause. IWQoS'20. https://doi.org/10.1109/IWQoS49365.2020.9213058
    [2] PCMCI. Science Advances'19. https://doi.org/10.1126/sciadv.aau4996
    """

    def __init__(
        self,
        alpha: float = 0.05,
        cond_ind_test: CondIndTest = None,
        tau_max: int = 3,
        max_conds_dim: int = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if cond_ind_test is None:
            cond_ind_test = _ParCorr(significance="analytic")
        self._alpha = alpha
        self._cond_ind_test = cond_ind_test
        self._tau_max = tau_max
        self._max_conds_dim = max_conds_dim

    @staticmethod
    def _gather_tau(p_matrix: np.ndarray) -> np.ndarray:
        # reason, result, tau
        num, result_num, _ = p_matrix.shape
        assert num == result_num

        link_matrix = []
        for reason in range(num):
            link_matrix.append([min(p_matrix[reason][result]) for result in range(num)])
        return np.array(link_matrix)

    def _create(
        self, data: np.ndarray, nodes: List[Node]
    ) -> Tuple[np.ndarray, List[Node]]:
        dataframe = data_processing.DataFrame(data)
        pcmci = PCMCI(
            dataframe=dataframe, cond_ind_test=self._cond_ind_test, verbosity=0
        )
        report = pcmci.run_pcmci(
            tau_max=self._tau_max,
            pc_alpha=self._alpha,
            max_conds_dim=self._max_conds_dim,
        )
        matrix = self._gather_tau(report["p_matrix"])
        return matrix, nodes



"""
Base GraphFactory based on data mining
"""
from typing import List
from typing import Tuple

import networkx as nx
import numpy as np

from .base import GraphFactory
from ..model.case import CaseData
from ..model.graph import Graph
from ..model.graph import Node
from ..model.graph import MemoryGraph


class EmptyGraphFactory(GraphFactory):
    """
    Create a graph with nodes only
    """

    def create(self, data: CaseData, current: float) -> Graph:
        graph = nx.DiGraph()
        graph.add_nodes_from(data.data_loader.nodes)
        return MemoryGraph(graph)


class StaticGraphFactory(GraphFactory):
    """
    Create the same graph
    """

    def __init__(self, graph: Graph, **kwargs):
        super().__init__(**kwargs)
        self._graph = graph

    def create(self, data: CaseData, current: float) -> Graph:
        return self._graph


class DynamicGraphFactory(GraphFactory):
    """
    Create graph mined from data
    """

    def _create(
        self, data: np.ndarray, nodes: List[Node]
    ) -> Tuple[np.ndarray, List[Node]]:
        """
        Mine graph from data

        Return:
            matrix: if matrix[i, j] is True, i may be one of the causes of j
            nodes: mapping from the matrix indexes to Nodes
        """
        raise NotImplementedError

    def create(self, data: CaseData, current: float) -> Graph:
        series = data.load_data(current=current)
        nodes = list(series.keys())
        series = np.array([series[node] for node in nodes]).T

        matrix, nodes = self._create(data=series, nodes=nodes)

        graph = nx.DiGraph()
        graph.add_nodes_from(nodes)
        graph.add_node(data.sli)
        graph.add_edges_from(
            (nodes[cause], nodes[effect]) for cause, effect in zip(*np.where(matrix))
        )
        return MemoryGraph(graph)



"""
Random walk
"""
from collections import defaultdict
from typing import Callable
from typing import Dict
from typing import List
from typing import Union

import numpy as np
import pandas as pd

from .base import Score
from .base import Scorer
from ..model.case import CaseData
from ..model.graph import Graph
from ..model.graph import Node


def _times(num: int, multiplier: int = 10) -> int:
    return num * multiplier


class RandomWalkScorer(Scorer):
    """
    Scorer based on random walk
    """

    def __init__(
        self,
        rho: float = 0.5,
        remove_sli: bool = False,
        num_loop: Union[int, Callable[[int], int]] = None,
        **kwargs,
    ):
        # pylint: disable=too-many-arguments
        super().__init__(**kwargs)
        self._rho = rho
        self._remove_sli = remove_sli
        self._num_loop = num_loop if num_loop is not None else _times
        self._rng = np.random.default_rng(self._seed)

    def generate_transition_matrix(
        self, graph: Graph, data: CaseData, scores: Dict[Node, Score]
    ) -> pd.DataFrame:
        """
        Generate the transition matrix
        """
        nodes = list(scores.keys())
        size = len(nodes)
        matrix = pd.DataFrame(np.zeros([size, size]), index=nodes, columns=nodes)
        for node in scores:
            for child in graph.children(node):
                if child in scores:
                    matrix[node][child] = self._rho * abs(scores[child].score)

            parents = graph.parents(node)
            if self._remove_sli:
                parents -= {data.sli}
            for parent in parents:
                if parent in scores:
                    matrix[node][parent] = abs(scores[parent].score)

            matrix[node][node] = max(abs(scores[node].score) - matrix[node].max(), 0)

            total_weight = matrix[node].sum()
            if total_weight > 0:
                matrix[node] /= total_weight
            else:
                matrix[node] = 1 / size
        return matrix

    def _walk(
        self, start: Node, num_loop: int, matrix: pd.DataFrame
    ) -> Dict[Node, int]:
        node: Node = start
        counter = defaultdict(int)
        for _ in range(num_loop):
            node = self._rng.choice(matrix.index, p=matrix[node])
            counter[node] += 1
        return counter

    def score(
        self,
        graph: Graph,
        data: CaseData,
        current: float,
        scores: Dict[Node, Score] = None,
    ) -> Dict[Node, Score]:
        if not scores:
            return scores

        matrix = self.generate_transition_matrix(graph=graph, data=data, scores=scores)
        if isinstance(self._num_loop, int):
            num_loop = self._num_loop
        else:
            num_loop = self._num_loop(len(scores))
        counter = self._walk(start=data.sli, num_loop=num_loop, matrix=matrix)

        for node, score in scores.items():
            score["pagerank"] = score.score = counter[node] / num_loop
        return scores


class SecondOrderRandomWalkScorer(RandomWalkScorer):
    """
    Scorer based on second-order random walk
    """

    def __init__(self, beta: float = 0.5, **kwargs):
        super().__init__(**kwargs)
        self._beta = beta

    def _walk(
        self, start: Node, num_loop: int, matrix: pd.DataFrame
    ) -> Dict[Node, int]:
        node: Node = start
        node_pre: Node = start
        counter = defaultdict(int)
        for _ in range(num_loop):
            prob_pre = matrix[node_pre][node]
            node_pre = node

            candidates: List[Node] = []
            weights: List[float] = []
            for key, value in matrix[node].iteritems():
                if value > 0:
                    candidates.append(key)
                    weights.append((1 - self._beta) * prob_pre + self._beta * value)
            total_weight = sum(weights)
            if total_weight == 0:
                node = self._rng.choice(candidates)
            else:
                node = self._rng.choice(
                    candidates, p=[weight / total_weight for weight in weights]
                )
            counter[node] += 1
        return counter



"""
This file is edited from `oracle/learn/mdn.py` in
    [the open-source code](https://github.com/perfd/perfd).

Reference:
[1] Silvery Fu, Saurabh Gupta, Radhika Mittal, and Sylvia Ratnasamy. On the Use of
    ML for Blackbox System Performance Prediction. NSDI 2021
"""
from typing import List
from typing import Tuple

import numpy as np
import torch
from torch import distributions as D
from torch import nn
from torch.nn import functional as F
from torch.utils.data import DataLoader
from torch.utils.data import TensorDataset
from torch.utils.data import random_split
import pytorch_lightning as pl
from pytorch_lightning.callbacks.early_stopping import EarlyStopping

from .base import GMM
from .base import GMMPredictor


class _MDN(nn.Module):
    def __init__(self, num_input, neurons: int = 50, components: int = 2, **kwargs):
        super().__init__(**kwargs)
        self._base = nn.Sequential(
            nn.Linear(num_input, neurons),
            nn.ReLU(),
            nn.Linear(neurons, neurons),
            nn.ReLU(),
        )
        self._alphas = nn.Linear(neurons, components)
        self._mus = nn.Linear(neurons, components)
        self._sigmas = nn.Linear(neurons, components)

    def forward(
        self, data: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Calculate model output
        """
        latent = self._base(data)
        alphas = F.softmax(self._alphas(latent), dim=-1)
        mus = self._mus(latent)
        sigmas = F.elu(self._sigmas(latent)) + 1 + 1e-6
        return alphas, mus, sigmas


class MDN(pl.LightningModule):
    # pylint: disable=abstract-method,arguments-differ,too-many-ancestors
    """
    Mixture Density Network
    """

    def __init__(self, learning_rate: float = 1e-3, **kwargs):
        super().__init__()
        self._model = _MDN(**kwargs)
        self._learning_rate = learning_rate

    def forward(
        self, data: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        return self._model(data)

    def loss(self, data: torch.Tensor, target: torch.Tensor):
        """
        Calculate loss for MDN
        """
        alphas, mus, sigmas = self.forward(data)
        comp = D.Normal(loc=mus, scale=sigmas)
        mix = D.Categorical(alphas)
        gmm = D.MixtureSameFamily(mix, comp)
        # ATTENTION: For multi-dimension output, we need transpose target
        likelihood = gmm.log_prob(target)
        return -likelihood.mean()

    def training_step(self, batch: Tuple[torch.Tensor, torch.Tensor], *args):
        data, target = batch
        loss = self.loss(data=data, target=target)
        self.log("train_loss", loss)
        return loss

    def validation_step(self, batch: Tuple[torch.Tensor, torch.Tensor], *args):
        data, target = batch
        loss = self.loss(data=data, target=target)
        self.log("val_loss", loss)
        return loss

    def configure_optimizers(self):
        return torch.optim.Adam(self._model.parameters(), lr=self._learning_rate)


class TorchGMM(GMM):
    """
    GMM with torch
    """

    def __init__(
        self,
        alphas: torch.Tensor,
        mus: torch.Tensor,
        sigmas: torch.Tensor,
    ):
        comp = D.Normal(loc=mus, scale=sigmas)
        mix = D.Categorical(alphas)
        gmm = D.MixtureSameFamily(mix, comp)
        self._model = gmm

    def sample(self, n_samples: int = 1) -> np.ndarray:
        return self._model.sample((n_samples,)).numpy()


class MDNPredictor(GMMPredictor):
    """
    Regress with a Mixture Density Network
    """

    DEFAULT_TRAIN_PARAMS = dict(
        logger=False,
        enable_checkpointing=False,
        enable_progress_bar=False,
        max_epochs=1000,
    )

    def __init__(
        self,
        batch_size: int = 128,
        train_ratio: float = 0.8,
        train_params: dict = None,
        model_params: dict = None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self._batch_size = batch_size
        self._train_ratio = train_ratio
        if train_params is None:
            train_params = {}
        self._train_params = {**self.DEFAULT_TRAIN_PARAMS, **train_params}
        self._model_params = model_params if model_params else {}
        self._model: MDN = None

    def train(self, train_x: np.ndarray, train_y: np.ndarray):
        """
        Train the model
        """
        num_samples, num_input = train_x.shape
        num_train = int(num_samples * self._train_ratio)
        data = TensorDataset(torch.FloatTensor(train_x), torch.FloatTensor(train_y))
        data_train, data_val = random_split(
            data,
            lengths=[num_train, num_samples - num_train],
            generator=torch.Generator().manual_seed(self._seed) if self._seed else None,
        )

        self._model = MDN(num_input=num_input, **self._model_params)
        trainer = pl.Trainer(
            callbacks=[
                EarlyStopping(monitor="val_loss", min_delta=0, patience=5, mode="min")
            ],
            **self._train_params,
        )
        trainer.fit(
            self._model,
            train_dataloaders=DataLoader(data_train),
            val_dataloaders=DataLoader(data_val),
        )

    def predict(self, test_x: np.ndarray) -> List[GMM]:
        """
        Generate GMM
        """
        data_test = DataLoader(
            TensorDataset(torch.FloatTensor(test_x)), batch_size=self._batch_size
        )
        gmms: List[GMM] = []
        for (data,) in data_test:
            output = self._model.forward(data)
            for alphas, mus, sigmas in zip(*[item.cpu() for item in output]):
                gmms.append(TorchGMM(alphas=alphas, mus=mus, sigmas=sigmas))
        return gmms
